#!/usr/bin/env python3
"""
Regularizers module.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

from abc import ABC, abstractmethod
from typing import Optional, Union
from collections.abc import Sequence

import numpy as np
import scipy.ndimage as spimg
from numpy.typing import DTypeLike, NDArray

from . import data_terms as dt
from . import operators

try:
    import pywt

    has_pywt = True
    use_swtn = pywt.version.version >= "1.0.2"  # type: ignore
except ImportError:
    has_pywt = False
    use_swtn = False
    print("WARNING - pywt was not found")


NDArrayInt = NDArray[np.integer]


# ---- Regularizers ----


class BaseRegularizer(ABC):
    """
    Initialize a base regularizer class, that defines the Regularizer object interface.

    Parameters
    ----------
    weight : Union[float, NDArray]
        The weight of the regularizer.
    norm : DataFidelityBase
        The norm of the regularizer minimization.
    """

    __reg_name__ = ""

    weight: NDArray
    dtype: DTypeLike
    op: Union[operators.BaseTransform, None]

    sigma: Union[float, NDArray]

    upd_mask: Optional[NDArray]

    def __init__(
        self,
        weight: Union[float, NDArray],
        norm: dt.DataFidelityBase,
        upd_mask: Optional[NDArray] = None,
        dtype: DTypeLike = np.float32,
    ):
        self.weight = np.array(weight)
        self.dtype = dtype
        self.op = None
        self.norm = norm
        self.upd_mask = upd_mask

    def info(self) -> str:
        """
        Return the regularizer info.

        Returns
        -------
        str
            Regularizer info string.
        """
        return self.__reg_name__ + f"(w:{self.weight.max():.3e})"

    def upper(self) -> str:
        """
        Return the upper case name of the regularizer.

        Returns
        -------
        str
            Upper case string name of the regularizer.
        """
        return self.__reg_name__.upper()

    def lower(self) -> str:
        """
        Return the lower case name of the regularizer.

        Returns
        -------
        str
            Lower case string name of the regularizer.
        """
        return self.__reg_name__.lower()

    @abstractmethod
    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        """
        Initialize the internal state, operator, and sigma. It then returns the tau.

        Parameters
        ----------
        primal : NDArray
            The primal vector.

        Returns
        -------
        Union[float, NDArray]
            The tau to be used in the SIRT or PDHG algorithm.
        """

    def initialize_dual(self) -> NDArray:
        """
        Return the initialized dual.

        Returns
        -------
        NDArray
            Initialized (zero) dual.
        """
        if self.op is None:
            raise ValueError("Regularizer not initialized! Please use method: `initialize_sigma_tau`.")

        return np.zeros(self.op.adj_shape, dtype=self.dtype)

    def update_dual(self, dual: NDArray, primal: NDArray) -> None:
        """
        Update the dual in-place.

        Parameters
        ----------
        dual : NDArray
            Current stat of the dual.
        primal : NDArray
            Primal or over-relaxation of the primal.
        """
        if self.op is None:
            raise ValueError("Regularizer not initialized! Please use method: `initialize_sigma_tau`.")

        dual += self.sigma * self.op(primal)

    def apply_proximal(self, dual: NDArray) -> None:
        """
        Apply the proximal operator to the dual in-place.

        Parameters
        ----------
        dual : NDArray
            The dual to be applied the proximal on.
        """
        if isinstance(self.norm, dt.DataFidelity_l1):
            self.norm.apply_proximal(dual, self.weight)
        else:
            self.norm.apply_proximal(dual)

    def compute_update_primal(self, dual: NDArray) -> NDArray:
        """
        Compute the partial update of a primal term, from this regularizer.

        Parameters
        ----------
        dual : NDArray
            The dual associated to this regularizer.

        Returns
        -------
        upd : NDArray
            The update to the primal.
        """
        if self.op is None:
            raise ValueError("Regularizer not initialized! Please use method: `initialize_sigma_tau`.")

        upd = self.op.T(dual)
        if self.upd_mask is not None:
            upd *= self.upd_mask
        if not isinstance(self.norm, dt.DataFidelity_l1):
            upd *= self.weight
        return upd

    def _check_primal(self, primal: NDArray) -> None:
        if self.dtype != primal.dtype:
            print(f"WARNING: Regularizer dtype ({self.dtype}) and primal dtype ({primal.dtype}) are different!")
            self.dtype = primal.dtype


class Regularizer_Grad(BaseRegularizer):
    """Gradient regularizer.

    When used with l1-norms, it promotes piece-wise constant reconstructions.
    When used with l2-norm, it promotes smooth reconstructions.

    Parameters
    ----------
    weight : Union[float, NDArray]
        The weight of the regularizer.
    ndims : int, optional
        The number of dimensions. The default is 2.
    axes : Sequence, optional
        The axes over which it computes the gradient. If None, it uses the last 2. The default is None.
    pad_mode: str, optional
        The padding mode to use for the linear convolution. The default is "edge".
    norm : DataFidelityBase, optional
        The norm of the regularizer minimization. The default is DataFidelity_l12().
    """

    __reg_name__ = "grad"

    def __init__(
        self,
        weight: Union[float, NDArray],
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l12(),
    ):
        super().__init__(weight=weight, norm=norm, upd_mask=upd_mask)

        if axes is None:
            axes = np.arange(-ndims, 0, dtype=int)
        elif not ndims == len(axes):
            print("WARNING - Number of axes different from number of dimensions. Updating dimensions accordingly.")
            ndims = len(axes)
        self.ndims = ndims
        self.axes = axes
        self.pad_mode = pad_mode.lower()

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        self._check_primal(primal)

        self.op = operators.TransformGradient(primal.shape, axes=self.axes, pad_mode=self.pad_mode)

        self.sigma = 0.5
        self.norm.assign_data(None, sigma=self.sigma)

        tau = 2 * self.ndims
        if not isinstance(self.norm, dt.DataFidelity_l1):
            tau *= self.weight
        if self.upd_mask is not None:
            tau = tau * self.upd_mask
        return tau


class Regularizer_TV1D(Regularizer_Grad):
    """Total Variation (TV) regularizer in 1D. It can be used to promote piece-wise constant reconstructions."""

    __reg_name__ = "TV1D"

    def __init__(
        self,
        weight: Union[float, NDArray],
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l12(),
    ):
        super().__init__(weight=weight, ndims=1, axes=axes, pad_mode=pad_mode, norm=norm, upd_mask=upd_mask)


class Regularizer_TV2D(Regularizer_Grad):
    """Total Variation (TV) regularizer in 2D. It can be used to promote piece-wise constant reconstructions."""

    __reg_name__ = "TV2D"

    def __init__(
        self,
        weight: Union[float, NDArray],
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l12(),
    ):
        super().__init__(weight=weight, ndims=2, axes=axes, pad_mode=pad_mode, norm=norm, upd_mask=upd_mask)


class Regularizer_TV3D(Regularizer_Grad):
    """Total Variation (TV) regularizer in 3D. It can be used to promote piece-wise constant reconstructions."""

    __reg_name__ = "TV3D"

    def __init__(
        self,
        weight: Union[float, NDArray],
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l12(),
    ):
        super().__init__(weight=weight, ndims=3, axes=axes, pad_mode=pad_mode, norm=norm, upd_mask=upd_mask)


class Regularizer_HubTV2D(Regularizer_Grad):
    """Total Variation (TV) regularizer in 2D. It can be used to promote piece-wise constant reconstructions."""

    __reg_name__ = "HubTV2D"

    def __init__(
        self,
        weight: Union[float, NDArray],
        huber_size: float,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
    ):
        super().__init__(
            weight=weight,
            ndims=2,
            axes=axes,
            pad_mode=pad_mode,
            upd_mask=upd_mask,
            norm=dt.DataFidelity_Huber(huber_size, l2_axis=0),
        )


class Regularizer_HubTV3D(Regularizer_Grad):
    """Total Variation (TV) regularizer in 3D. It can be used to promote piece-wise constant reconstructions."""

    __reg_name__ = "HubTV3D"

    def __init__(
        self,
        weight: Union[float, NDArray],
        huber_size: float,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
    ):
        super().__init__(
            weight=weight,
            ndims=3,
            axes=axes,
            pad_mode=pad_mode,
            upd_mask=upd_mask,
            norm=dt.DataFidelity_Huber(huber_size, l2_axis=0),
        )


class Regularizer_smooth1D(Regularizer_Grad):
    """It can be used to promote smooth reconstructions."""

    __reg_name__ = "smooth1D"

    def __init__(
        self,
        weight: Union[float, NDArray],
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l2(),
    ):
        super().__init__(weight=weight, ndims=1, axes=axes, pad_mode=pad_mode, norm=norm, upd_mask=upd_mask)


class Regularizer_smooth2D(Regularizer_Grad):
    """It can be used to promote smooth reconstructions."""

    __reg_name__ = "smooth2D"

    def __init__(
        self,
        weight: Union[float, NDArray],
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l2(),
    ):
        super().__init__(weight=weight, ndims=2, axes=axes, pad_mode=pad_mode, norm=norm, upd_mask=upd_mask)


class Regularizer_smooth3D(Regularizer_Grad):
    """It can be used to promote smooth reconstructions."""

    __reg_name__ = "smooth3D"

    def __init__(
        self,
        weight: Union[float, NDArray],
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l2(),
    ):
        super().__init__(weight=weight, ndims=3, axes=axes, pad_mode=pad_mode, norm=norm, upd_mask=upd_mask)


class Regularizer_lap(BaseRegularizer):
    """Laplacian regularizer. It can be used to promote smooth reconstructions."""

    __reg_name__ = "lap"

    def __init__(
        self,
        weight: Union[float, NDArray],
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
    ):
        super().__init__(weight=weight, norm=dt.DataFidelity_l1(), upd_mask=upd_mask)

        if axes is None:
            axes = np.arange(-ndims, 0, dtype=int)
        elif not ndims == len(axes):
            print("WARNING - Number of axes different from number of dimensions. Updating dimensions accordingly.")
            ndims = len(axes)
        self.ndims = ndims
        self.axes = axes
        self.pad_mode = pad_mode.lower()

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        self._check_primal(primal)

        self.op = operators.TransformLaplacian(primal.shape, axes=self.axes, pad_mode=self.pad_mode)

        self.sigma = 0.25
        self.norm.assign_data(None, sigma=self.sigma)

        tau = 4 * self.ndims
        if self.upd_mask is not None:
            tau = tau * self.upd_mask
        return tau


class Regularizer_lap1D(Regularizer_lap):
    """Laplacian regularizer in 1D. It can be used to promote smooth reconstructions."""

    __reg_name__ = "lap1D"

    def __init__(
        self,
        weight,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
    ):
        Regularizer_lap.__init__(self, weight=weight, ndims=1, axes=axes, pad_mode=pad_mode, upd_mask=upd_mask)


class Regularizer_lap2D(Regularizer_lap):
    """Laplacian regularizer in 2D. It can be used to promote smooth reconstructions."""

    __reg_name__ = "lap2D"

    def __init__(
        self,
        weight,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
    ):
        Regularizer_lap.__init__(self, weight=weight, ndims=2, axes=axes, pad_mode=pad_mode, upd_mask=upd_mask)


class Regularizer_lap3D(Regularizer_lap):
    """Laplacian regularizer in 3D. It can be used to promote smooth reconstructions."""

    __reg_name__ = "lap3D"

    def __init__(
        self,
        weight,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
    ):
        Regularizer_lap.__init__(self, weight=weight, ndims=3, axes=axes, pad_mode=pad_mode, upd_mask=upd_mask)


class Regularizer_l1(BaseRegularizer):
    """l1-norm regularizer. It can be used to promote sparse reconstructions."""

    __reg_name__ = "l1"

    def __init__(
        self,
        weight: Union[float, NDArray],
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l1(),
    ):
        super().__init__(weight=weight, norm=norm, upd_mask=upd_mask)

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        self._check_primal(primal)

        self.op = operators.TransformIdentity(primal.shape)

        self.norm.assign_data(None, sigma=1)

        tau = 1.0
        if self.upd_mask is not None:
            tau = tau * self.upd_mask
        return tau

    def update_dual(self, dual: NDArray, primal: NDArray) -> None:
        dual += primal


class Regularizer_swl(BaseRegularizer):
    """Base stationary wavelet regularizer. It can be used to promote sparse reconstructions in the wavelet domain."""

    __reg_name__ = "swl"

    def info(self) -> str:
        """
        Return the regularizer info.

        Returns
        -------
        str
            Regularizer info string.
        """
        return self.__reg_name__ + "(t:" + self.wavelet + "-l:%d" % self.level + "-w:%g" % self.weight.max() + ")"

    def __init__(
        self,
        weight: Union[float, NDArray],
        wavelet: str,
        level: int,
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_on_demand: str = "constant",
        upd_mask: Optional[NDArray] = None,
        normalized: bool = False,
        min_approx: bool = True,
        norm: dt.DataFidelityBase = dt.DataFidelity_l1(),
    ):
        if not has_pywt:
            raise ValueError("Cannot use wavelet regularizer because pywavelets is not installed.")
        if not use_swtn:
            raise ValueError("Cannot use stationary wavelet regularizer because pywavelets is too old (<1.0.2).")
        super().__init__(weight=weight, norm=norm, upd_mask=upd_mask)
        self.wavelet = wavelet
        self.level = level
        self.normalized = normalized
        self.min_approx = min_approx

        if axes is None:
            axes = np.arange(-ndims, 0, dtype=int)
        elif not ndims == len(axes):
            print("WARNING - Number of axes different from number of dimensions. Updating dimensions accordingly.")
            ndims = len(axes)
        self.ndims = ndims
        self.axes = axes

        self.pad_on_demand = pad_on_demand

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        self._check_primal(primal)

        self.op = operators.TransformStationaryWavelet(
            primal.shape,
            wavelet=self.wavelet,
            level=self.level,
            axes=self.axes,
            pad_on_demand=self.pad_on_demand,
            normalized=self.normalized,
        )

        filt_bank_l1norm = np.linalg.norm(self.op.w.filter_bank, ord=1, axis=-1)
        lo_dec_mult = filt_bank_l1norm[0] ** self.ndims
        lo_rec_mult = filt_bank_l1norm[2] ** self.ndims

        self.dec_func_mult = (
            self.op.wlet_dec_filter_mult[None, :] * (lo_dec_mult ** np.arange(self.level - 1, -1, -1))[:, None]
        )
        self.dec_func_mult = np.concatenate(([lo_dec_mult**self.level], self.dec_func_mult.flatten()))

        self.rec_func_mult = (
            self.op.wlet_rec_filter_mult[None, :] * (lo_rec_mult ** np.arange(self.level - 1, -1, -1))[:, None]
        )
        self.rec_func_mult = np.concatenate(([lo_rec_mult**self.level], self.rec_func_mult.flatten()))

        # self.dec_func_mult = 2 ** np.arange(self.level, 0, -1)
        # self.dec_func_mult = np.tile(self.dec_func_mult[:, None], [1, (2 ** self.ndims) - 1])
        # self.dec_func_mult = np.concatenate(([self.dec_func_mult[0, 0]], self.dec_func_mult.flatten()))

        if self.normalized:
            self.sigma = 1
            self.norm.assign_data(None, sigma=self.sigma)

            tau = self.dec_func_mult.size
        else:
            self.sigma = np.reshape(1 / self.dec_func_mult, [-1] + [1] * len(self.op.dir_shape))
            self.norm.assign_data(None, sigma=self.sigma)

            tau = np.ones_like(self.rec_func_mult) * ((2**self.ndims) - 1)
            tau[0] += 1
            tau = np.sum(tau / self.rec_func_mult)

        if not isinstance(self.norm, dt.DataFidelity_l1):
            tau *= self.weight
        if self.upd_mask is not None:
            tau = tau * self.upd_mask
        return tau

    def update_dual(self, dual: NDArray, primal: NDArray) -> None:
        if self.op is None:
            raise ValueError("Regularizer not initialized! Please use method: `initialize_sigma_tau`.")

        upd = self.op(primal)
        if not self.normalized:
            upd *= self.sigma
        dual += upd
        if not self.min_approx:
            dual[0, ...] = 0

    def apply_proximal(self, dual: NDArray) -> None:
        if isinstance(self.norm, dt.DataFidelity_l12):
            tmp_dual = dual[1:]
            tmp_dual = tmp_dual.reshape([-1, self.level, *dual.shape[1:]])
            self.norm.apply_proximal(tmp_dual, self.weight)
            tmp_dual = dual[0:1:]
            self.norm.apply_proximal(tmp_dual, self.weight)
        else:
            super().apply_proximal(dual)


class Regularizer_l1swl(Regularizer_swl):
    """l1-norm Wavelet regularizer. It can be used to promote sparse reconstructions in the wavelet domain."""

    __reg_name__ = "l1swl"

    def __init__(
        self,
        weight: Union[float, NDArray],
        wavelet: str,
        level: int,
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_on_demand: str = "constant",
        upd_mask: Optional[NDArray] = None,
        normalized: bool = False,
        min_approx: bool = True,
    ):
        super().__init__(
            weight,
            wavelet,
            level,
            ndims=ndims,
            axes=axes,
            pad_on_demand=pad_on_demand,
            upd_mask=upd_mask,
            normalized=normalized,
            min_approx=min_approx,
            norm=dt.DataFidelity_l1(),
        )


class Regularizer_l12swl(Regularizer_swl):
    """l1-norm Wavelet regularizer. It can be used to promote sparse reconstructions in the wavelet domain."""

    __reg_name__ = "l12swl"

    def __init__(
        self,
        weight: Union[float, NDArray],
        wavelet: str,
        level: int,
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_on_demand: str = "constant",
        upd_mask: Optional[NDArray] = None,
        normalized: bool = False,
        min_approx: bool = True,
    ):
        super().__init__(
            weight,
            wavelet,
            level,
            ndims=ndims,
            axes=axes,
            pad_on_demand=pad_on_demand,
            upd_mask=upd_mask,
            normalized=normalized,
            min_approx=min_approx,
            norm=dt.DataFidelity_l12(),
        )


class Regularizer_Hub_swl(Regularizer_swl):
    """l1-norm Wavelet regularizer. It can be used to promote sparse reconstructions in the wavelet domain."""

    __reg_name__ = "Hubswl"

    def __init__(
        self,
        weight: Union[float, NDArray],
        wavelet: str,
        level: int,
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_on_demand: str = "constant",
        upd_mask: Optional[NDArray] = None,
        normalized: bool = False,
        min_approx: bool = True,
        huber_size: Optional[int] = None,
    ):
        super().__init__(
            weight,
            wavelet,
            level,
            ndims=ndims,
            axes=axes,
            pad_on_demand=pad_on_demand,
            upd_mask=upd_mask,
            normalized=normalized,
            min_approx=min_approx,
            norm=dt.DataFidelity_Huber(huber_size),
        )


class Regularizer_dwl(BaseRegularizer):
    """Base decimated wavelet regularizer. It can be used to promote sparse reconstructions in the wavelet domain."""

    __reg_name__ = "dwl"

    def info(self) -> str:
        """
        Return the regularizer info.

        Returns
        -------
        str
            Regularizer info string.
        """
        return self.__reg_name__ + "(t:" + self.wavelet + "-l:%d" % self.level + "-w:%g" % self.weight.max() + ")"

    def __init__(
        self,
        weight: Union[float, NDArray],
        wavelet: str,
        level: int,
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_on_demand: str = "constant",
        upd_mask: Optional[NDArray] = None,
        min_approx: bool = True,
        norm: dt.DataFidelityBase = dt.DataFidelity_l1(),
    ):
        if not has_pywt:
            raise ValueError("Cannot use wavelet regularizer because pywavelets is not installed.")
        super().__init__(weight=weight, norm=norm, upd_mask=upd_mask)
        self.wavelet = wavelet
        self.level = level
        self.min_approx = min_approx

        if axes is None:
            axes = np.arange(-ndims, 0, dtype=int)
        elif not ndims == len(axes):
            print("WARNING - Number of axes different from number of dimensions. Updating dimensions accordingly.")
            ndims = len(axes)
        self.ndims = ndims
        self.axes = axes

        self.pad_on_demand = pad_on_demand

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        self._check_primal(primal)

        self.op = operators.TransformDecimatedWavelet(
            primal.shape, wavelet=self.wavelet, level=self.level, axes=self.axes, pad_on_demand=self.pad_on_demand
        )

        filt_bank_l1norm = np.linalg.norm(self.op.w.filter_bank, ord=1, axis=-1)
        lo_dec_mult = filt_bank_l1norm[0] ** self.ndims
        lo_rec_mult = filt_bank_l1norm[2] ** self.ndims

        self.dec_func_mult = (
            self.op.wlet_dec_filter_mult[None, :] * (lo_dec_mult ** np.arange(self.level - 1, -1, -1))[:, None]
        )
        self.dec_func_mult = np.concatenate(([lo_dec_mult**self.level], self.dec_func_mult.flatten()))

        self.rec_func_mult = (
            self.op.wlet_rec_filter_mult[None, :] * (lo_rec_mult ** np.arange(self.level - 1, -1, -1))[:, None]
        )
        self.rec_func_mult = np.concatenate(([lo_rec_mult**self.level], self.rec_func_mult.flatten()))

        # self.dec_func_mult = 2 ** np.arange(self.level, 0, -1)
        # self.rec_func_mult = self.dec_func_mult

        tmp_sigma = [np.ones(self.op.sub_band_shapes[0], self.dtype) * self.dec_func_mult[0]]
        count = 0
        for ii_l in range(self.level):
            d = {}
            for label in self.op.sub_band_shapes[ii_l + 1].keys():
                # d[label] = np.ones(self.op.sub_band_shapes[ii_l + 1][label], self.dtype) * self.dec_func_mult[ii_l]
                d[label] = np.ones(self.op.sub_band_shapes[ii_l + 1][label], self.dtype) * self.dec_func_mult[count]
                count += 1
            tmp_sigma.append(d)
        self.sigma, _ = pywt.coeffs_to_array(tmp_sigma, axes=self.axes)
        self.norm.assign_data(None, sigma=self.sigma)

        tau = np.ones_like(self.rec_func_mult) * ((2**self.ndims) - 1)
        tau[0] += 1
        tau = np.sum(tau / self.rec_func_mult)

        if not isinstance(self.norm, dt.DataFidelity_l1):
            tau *= self.weight
        if self.upd_mask is not None:
            tau = tau * self.upd_mask
        return tau

    def update_dual(self, dual: NDArray, primal: NDArray) -> None:
        super().update_dual(dual, primal)
        if not self.min_approx:
            if self.op is None:
                raise ValueError("Regularizer not initialized! Please use method: `initialize_sigma_tau`.")

            op_wl: operators.TransformDecimatedWavelet = self.op  # type: ignore
            slices = [slice(0, x) for x in op_wl.sub_band_shapes[0]]
            dual[tuple(slices)] = 0

    def apply_proximal(self, dual: NDArray) -> None:
        if isinstance(self.norm, dt.DataFidelity_l12):
            op_wl: operators.TransformDecimatedWavelet = self.op
            coeffs = pywt.array_to_coeffs(dual, op_wl.slicing_info)
            for ii_l in range(1, len(coeffs)):
                c_l = coeffs[ii_l]
                labels = []
                details = []
                for lab, det in c_l.items():
                    labels.append(lab)
                    details.append(det)
                c_ll = np.stack(details, axis=0)
                self.norm.apply_proximal(c_ll, self.weight)
                for ii, lab in enumerate(labels):
                    c_l[lab] = c_ll[ii]
                coeffs[ii_l] = c_l
            self.norm.apply_proximal(coeffs[0], self.weight)
            dual[:] = pywt.coeffs_to_array(coeffs)[0]
        else:
            super().apply_proximal(dual)


class Regularizer_l1dwl(Regularizer_dwl):
    """l1-norm decimated wavelet regularizer. It can be used to promote sparse reconstructions."""

    __reg_name__ = "l1dwl"

    def __init__(
        self,
        weight: Union[float, NDArray],
        wavelet: str,
        level: int,
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_on_demand: str = "constant",
        upd_mask: Optional[NDArray] = None,
    ):
        super().__init__(
            weight,
            wavelet,
            level,
            ndims=ndims,
            axes=axes,
            pad_on_demand=pad_on_demand,
            upd_mask=upd_mask,
            norm=dt.DataFidelity_l1(),
        )


class Regularizer_l12dwl(Regularizer_dwl):
    """l1-norm decimated wavelet regularizer. It can be used to promote sparse reconstructions."""

    __reg_name__ = "l12dwl"

    def __init__(
        self,
        weight: Union[float, NDArray],
        wavelet: str,
        level: int,
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_on_demand: str = "constant",
        upd_mask: Optional[NDArray] = None,
    ):
        super().__init__(
            weight,
            wavelet,
            level,
            ndims=ndims,
            axes=axes,
            pad_on_demand=pad_on_demand,
            upd_mask=upd_mask,
            norm=dt.DataFidelity_l12(),
        )


class Regularizer_Hub_dwl(Regularizer_dwl):
    """l1-norm decimated wavelet regularizer. It can be used to promote sparse reconstructions."""

    __reg_name__ = "Hubdwl"

    def __init__(
        self,
        weight: Union[float, NDArray],
        wavelet: str,
        level: int,
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_on_demand: str = "constant",
        upd_mask: Optional[NDArray] = None,
        huber_size: Optional[int] = None,
    ):
        super().__init__(
            weight,
            wavelet,
            level,
            ndims=ndims,
            axes=axes,
            pad_on_demand=pad_on_demand,
            upd_mask=upd_mask,
            norm=dt.DataFidelity_Huber(huber_size),
        )


class BaseRegularizer_med(BaseRegularizer):
    """Median filter regularizer base class. It can be used to promote filtered reconstructions."""

    __reg_name__ = "med"

    def info(self) -> str:
        """
        Return the regularizer info.

        Returns
        -------
        str
            Regularizer info string.
        """
        return self.__reg_name__ + "(s:%s" % np.array(self.filt_size) + "-w:%g" % self.weight.max() + ")"

    def __init__(
        self,
        weight: Union[float, NDArray],
        filt_size: int = 3,
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l1(),
    ):
        super().__init__(weight=weight, norm=norm, upd_mask=upd_mask)
        self.filt_size = filt_size

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        self._check_primal(primal)

        self.op = operators.TransformIdentity(primal.shape)
        self.norm.assign_data(None, sigma=1)

        if not isinstance(self.norm, dt.DataFidelity_l1):
            tau = self.weight
        else:
            tau = 1.0
        if self.upd_mask is not None:
            tau = tau * self.upd_mask
        return tau

    def update_dual(self, dual: NDArray, primal: NDArray) -> None:
        dual += primal - spimg.median_filter(primal, self.filt_size)


class Regularizer_l1med(BaseRegularizer_med):
    """l1-norm median filter regularizer. It can be used to promote filtered reconstructions."""

    __reg_name__ = "l1med"

    def __init__(self, weight: Union[float, NDArray], filt_size: int = 3):
        BaseRegularizer_med.__init__(self, weight, filt_size=filt_size, norm=dt.DataFidelity_l1())


class Regularizer_l2med(BaseRegularizer_med):
    """l2-norm median filter regularizer. It can be used to promote filtered reconstructions."""

    __reg_name__ = "l2med"

    def __init__(self, weight: Union[float, NDArray], filt_size: int = 3):
        BaseRegularizer_med.__init__(self, weight, filt_size=filt_size, norm=dt.DataFidelity_l2())


class Regularizer_fft(BaseRegularizer):
    """Fourier regularizer. It can be used to promote sparse reconstructions in the Fourier domain."""

    __reg_name__ = "fft"

    def __init__(
        self,
        weight: Union[float, NDArray],
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        fft_filter: str = "exp",
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l12(),
    ):
        super().__init__(weight=weight, norm=norm, upd_mask=upd_mask)

        if axes is None:
            axes = np.arange(-ndims, 0, dtype=int)
        elif not ndims == len(axes):
            print("WARNING - Number of axes different from number of dimensions. Updating dimensions accordingly.")
            ndims = len(axes)
        self.ndims = ndims
        self.axes = axes

        self.fft_filter = fft_filter

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        self._check_primal(primal)

        self.op = operators.TransformFourier(primal.shape, axes=self.axes)

        if isinstance(self.fft_filter, str):
            coords = [np.fft.fftfreq(s) for s in self.op.adj_shape[self.axes]]
            coords = np.array(np.meshgrid(*coords, indexing="ij"))

            if self.fft_filter.lower() == "delta":
                self.sigma = 1 - np.all(coords == 0, axis=0)
            elif self.fft_filter.lower() == "exp":
                self.sigma = 1 - np.exp(-np.sqrt(np.sum(coords**2, axis=0)) * 12)
            elif self.fft_filter.lower() == "exp2":
                self.sigma = 1 - np.exp(-np.sum(coords**2, axis=0) * 36)
            else:
                raise ValueError('Unknown FFT mask: %s. Options are: "delta", "exp". and "exp2".' % self.fft_filter)

            new_shape = np.ones_like(self.op.adj_shape)
            new_shape[self.axes] = self.op.adj_shape[self.axes]
            self.sigma = np.reshape(self.sigma, new_shape)
        else:
            self.sigma = 1

        self.norm.assign_data(None, sigma=self.sigma)

        if not isinstance(self.norm, dt.DataFidelity_l1):
            tau = self.weight
        else:
            tau = 1.0
        if self.upd_mask is not None:
            tau = tau * self.upd_mask
        return tau


# Multi-channel regularizers


class Regularizer_TNV(Regularizer_Grad):
    """Total Nuclear Variation (TNV) regularizer.

    It can be used to promote piece-wise constant reconstructions, for multi-channel volumes.
    """

    __reg_name__ = "TNV"

    def __init__(
        self,
        weight: Union[float, NDArray],
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_mode: str = "edge",
        upd_mask: Optional[NDArray] = None,
        spectral_norm: dt.DataFidelityBase = dt.DataFidelity_l1(),
        x_ref: Optional[NDArray] = None,
    ):
        super().__init__(weight=weight, ndims=ndims, axes=axes, pad_mode=pad_mode, upd_mask=upd_mask)

        # Here we assume that the channels will be the rows and the derivatives the columns
        self.norm = dt.DataFidelity_ln(ln_axes=(1, 0), spectral_norm=spectral_norm)
        self.x_ref = x_ref

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        tau = super().initialize_sigma_tau(primal)

        if self.x_ref is not None:
            if self.op is None:
                raise ValueError("Regularizer should have been initialized... this is a bug!")

            self.q_ref = self.op(self.x_ref)
            self.q_ref = np.expand_dims(self.q_ref, axis=1)

        return tau


class Regularizer_VTV(Regularizer_Grad):
    """Vectorial Total Variation (VTV) regularizer.

    It can be used to promote piece-wise constant reconstructions, for multi-channel volumes.
    """

    __reg_name__ = "VTV"

    def __init__(
        self,
        weight: Union[float, NDArray],
        ndims: int = 2,
        pwise_der_norm: Union[int, float] = 2,
        pwise_chan_norm: Union[int, float] = np.inf,
        x_ref: Optional[NDArray] = None,
        upd_mask: Optional[NDArray] = None,
    ):
        super().__init__(weight=weight, ndims=ndims, upd_mask=upd_mask)

        self.pwise_der_norm = pwise_der_norm
        if self.pwise_der_norm not in [1, 2, np.inf]:
            self._raise_pwise_norm_error()

        self.pwise_chan_norm = pwise_chan_norm
        if self.pwise_chan_norm not in [1, 2, np.inf]:
            self._raise_pwise_norm_error()

        if x_ref is not None:
            self.initialize_sigma_tau(x_ref)
            q_ref = self.initialize_dual()
            self.update_dual(q_ref, x_ref)
            self.q_ref = np.expand_dims(q_ref, axis=1)
        else:
            self.q_ref = None

    def _raise_pwise_norm_error(self):
        raise ValueError(
            "The only supported point-wise norm exponents are: 1, 2, and Inf."
            + f" Provided the following instead: derivatives={self.pwise_der_norm}, channel={self.pwise_chan_norm}"
        )

    def apply_proximal(self, dual: NDArray) -> None:
        # Following assignments will detach the local array from the original one
        dual_tmp = dual.copy()

        dual_is_scalar = len(dual_tmp.shape) == (self.ndims + 1)
        if dual_is_scalar:
            dual_tmp = np.expand_dims(dual_tmp, axis=1)

        if self.q_ref is not None:
            dual_tmp = np.concatenate((dual_tmp, self.q_ref), axis=1)

        if self.pwise_der_norm == 1:
            grad_norm = np.abs(dual_tmp)
        elif self.pwise_der_norm == 2:
            grad_norm = np.linalg.norm(dual_tmp, axis=0, ord=2, keepdims=True)
        elif self.pwise_der_norm == np.inf:
            grad_norm = np.linalg.norm(dual_tmp, axis=0, ord=1, keepdims=True)
        else:
            self._raise_pwise_norm_error()

        if self.pwise_chan_norm == 1:
            dual_norm = grad_norm
        elif self.pwise_chan_norm == 2:
            dual_norm = np.linalg.norm(grad_norm, axis=1, ord=2, keepdims=True)
        elif self.pwise_chan_norm == np.inf:
            dual_norm = np.linalg.norm(grad_norm, axis=1, ord=1, keepdims=True)
        else:
            self._raise_pwise_norm_error()

        dual_tmp /= np.fmax(dual_norm, self.weight)
        dual_tmp *= self.weight

        if self.q_ref is not None:
            dual_tmp = dual_tmp[:, : dual_tmp.shape[1] - 1 :, ...]

        if dual_is_scalar:
            dual_tmp = np.squeeze(dual_tmp, axis=1)

        dual[:] = dual_tmp[:]  # Replacing values


class Regularizer_lnswl(Regularizer_l1swl):
    """Nuclear-norm Wavelet regularizer.

    It can be used to promote compressed multi-channel reconstructions.
    """

    __reg_name__ = "lnswl"

    def __init__(
        self,
        weight: Union[float, NDArray],
        wavelet: str,
        level: int,
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_on_demand: str = "constant",
        upd_mask: Optional[NDArray] = None,
        normalized: bool = False,
        min_approx: bool = True,
        spectral_norm: dt.DataFidelityBase = dt.DataFidelity_l1(),
        x_ref: Optional[NDArray] = None,
    ):
        super().__init__(
            weight,
            wavelet,
            level,
            ndims=ndims,
            axes=axes,
            pad_on_demand=pad_on_demand,
            upd_mask=upd_mask,
            normalized=normalized,
            min_approx=min_approx,
        )

        self.norm = dt.DataFidelity_ln(ln_axes=(1, 0), spectral_norm=spectral_norm)
        self.x_ref = x_ref

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        tau = super().initialize_sigma_tau(primal)

        if self.x_ref is not None:
            if self.op is None:
                raise ValueError("Regularizer should have been initialized... this is a bug!")

            self.q_ref = self.op(self.x_ref)
            self.q_ref = np.expand_dims(self.q_ref, axis=1)

        return tau


class Regularizer_vl1wl(Regularizer_l1swl):
    """l1-norm vectorial Wavelet regularizer. It can be used to promote compressed reconstructions."""

    __reg_name__ = "vl1wl"

    def __init__(
        self,
        weight: Union[float, NDArray],
        wavelet: str,
        level: int,
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        pad_on_demand: str = "constant",
        upd_mask: Optional[NDArray] = None,
        normalized: bool = False,
        min_approx: bool = True,
        pwise_lvl_norm: Union[int, float] = 1,
        pwise_chan_norm: Union[int, float] = np.inf,
        x_ref: Optional[NDArray] = None,
    ):
        super().__init__(
            weight,
            wavelet,
            level,
            ndims=ndims,
            axes=axes,
            pad_on_demand=pad_on_demand,
            upd_mask=upd_mask,
            normalized=normalized,
            min_approx=min_approx,
        )

        self.pwise_lvl_norm = pwise_lvl_norm
        if self.pwise_lvl_norm not in [1, 2, np.inf]:
            self._raise_pwise_norm_error()

        self.pwise_chan_norm = pwise_chan_norm
        if self.pwise_chan_norm not in [1, 2, np.inf]:
            self._raise_pwise_norm_error()

        self.x_ref = x_ref
        self.q_ref = None

    def _raise_pwise_norm_error(self):
        raise ValueError(
            "The only supported point-wise norm exponents are: 1, 2, and Inf."
            + f" Provided the following instead: level={self.pwise_lvl_norm}, channel={self.pwise_chan_norm}"
        )

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        tau = super().initialize_sigma_tau(primal)

        if self.x_ref is not None:
            if self.op is None:
                raise ValueError("Regularizer should have been initialized... this is a bug!")

            self.q_ref = self.op(self.x_ref)

        return tau

    def apply_proximal(self, dual: NDArray) -> None:
        dual_tmp = dual.copy()

        if self.q_ref is not None:
            dual_tmp = np.concatenate((dual_tmp, self.q_ref), axis=1)

        if self.pwise_lvl_norm == 1:
            lvl_norm = np.abs(dual_tmp)
        elif self.pwise_lvl_norm == 2:
            lvl_norm = np.linalg.norm(dual_tmp, axis=0, ord=2, keepdims=True)
        elif self.pwise_lvl_norm == np.inf:
            lvl_norm = np.linalg.norm(dual_tmp, axis=0, ord=1, keepdims=True)
        else:
            self._raise_pwise_norm_error()

        if self.pwise_chan_norm == 1:
            dual_norm = lvl_norm
        elif self.pwise_chan_norm == 2:
            dual_norm = np.linalg.norm(lvl_norm, axis=1, ord=2, keepdims=True)
        elif self.pwise_chan_norm == np.inf:
            dual_norm = np.linalg.norm(lvl_norm, axis=1, ord=1, keepdims=True)
        else:
            self._raise_pwise_norm_error()

        dual_tmp /= np.fmax(dual_norm, self.weight)
        dual_tmp *= self.weight

        if self.q_ref is not None:
            dual_tmp = dual_tmp[:, : dual_tmp.shape[0] - 1 :, ...]

        dual[:] = dual_tmp[:]


class Regularizer_vSVD(BaseRegularizer):
    """Regularizer based on the Singular Value Decomposition.

    It can be used to promote similar reconstructions across different channels.
    """

    __reg_name__ = "vsvd"

    def __init__(
        self,
        weight: Union[float, NDArray],
        ndims: int = 2,
        axes: Union[Sequence[int], NDArray, None] = None,
        axis_channels: Sequence[int] = (0,),
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l1(),
    ):
        super().__init__(weight=weight, norm=norm, upd_mask=upd_mask)

        if axes is None:
            axes = np.arange(-ndims, 0, dtype=int)
        elif not ndims == len(axes):
            print("WARNING - Number of axes different from number of dimensions. Updating dimensions accordingly.")
            ndims = len(axes)
        self.ndims = ndims
        self.axes = np.array(axes)
        self.axis_channels = axis_channels

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        self._check_primal(primal)

        self.op = operators.TransformSVD(primal.shape, axes_rows=self.axis_channels, axes_cols=self.axes, rescale=True)

        self.sigma = 1
        self.norm.assign_data(None, sigma=self.sigma)

        if not isinstance(self.norm, dt.DataFidelity_l1):
            tau = self.weight
        else:
            tau = 1.0
        if self.upd_mask is not None:
            tau = tau * self.upd_mask
        return tau


# ---- Constraints ----


class Constraint_LowerLimit(BaseRegularizer):
    """Lower limit constraint. It can be used to promote reconstructions in certain regions of solution space."""

    __reg_name__ = "lowlim"

    def info(self) -> str:
        """
        Return the regularizer info.

        Returns
        -------
        str
            Regularizer info string.
        """
        return self.__reg_name__ + "(l:%g" % self.limit + ")"

    def __init__(
        self,
        limit: Union[float, NDArray],
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l2(),
    ):
        super().__init__(weight=1, norm=norm, upd_mask=upd_mask)
        self.limit = limit

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        self._check_primal(primal)

        self.op = operators.TransformIdentity(primal.shape)

        self.norm.assign_data(self.limit, sigma=1)

        if not isinstance(self.norm, dt.DataFidelity_l1):
            tau = self.weight
        else:
            tau = 1.0
        if self.upd_mask is not None:
            tau = tau * self.upd_mask
        return tau

    def update_dual(self, dual: NDArray, primal: NDArray) -> None:
        dual += primal

    def apply_proximal(self, dual: NDArray) -> None:
        dual[dual > self.limit] = self.limit
        self.norm.apply_proximal(dual)


class Constraint_UpperLimit(BaseRegularizer):
    """Upper limit constraint. It can be used to promote reconstructions in certain regions of solution space."""

    __reg_name__ = "uplim"

    def info(self) -> str:
        """
        Return the regularizer info.

        Returns
        -------
        str
            Regularizer info string.
        """
        return self.__reg_name__ + "(l:%g" % self.limit + ")"

    def __init__(
        self,
        limit: Union[float, NDArray],
        upd_mask: Optional[NDArray] = None,
        norm: dt.DataFidelityBase = dt.DataFidelity_l2(),
    ):
        super().__init__(weight=1, norm=norm, upd_mask=upd_mask)
        self.limit = limit

    def initialize_sigma_tau(self, primal: NDArray) -> Union[float, NDArray]:
        self._check_primal(primal)

        self.op = operators.TransformIdentity(primal.shape)

        self.norm.assign_data(self.limit, sigma=1)

        if not isinstance(self.norm, dt.DataFidelity_l1):
            tau = self.weight
        else:
            tau = 1.0
        if self.upd_mask is not None:
            tau = tau * self.upd_mask
        return tau

    def update_dual(self, dual: NDArray, primal: NDArray) -> None:
        dual += primal

    def apply_proximal(self, dual: NDArray) -> None:
        dual[dual < self.limit] = self.limit
        self.norm.apply_proximal(dual)



#!/usr/bin/env python

"""
Test `corrct.solvers` package.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import numpy as np
import pytest
from numpy.typing import NDArray
from corrct import processing, projectors, solvers, regularizers


eps = np.finfo(np.float32).eps
OVERSIZE_DATA = 5

VOL_SHAPE_2D = np.asarray((10, 10))
PRJ_SHAPE_2D = np.asarray((OVERSIZE_DATA, np.prod(VOL_SHAPE_2D)))

TOLERANCE = 1e-2


def _print_max_deviation(test_ind: str, sol_diff: NDArray[np.floating], tolerance: float) -> None:
    """
    Print the maximum absolute deviation between the solution and the reference volume.

    Parameters
    ----------
    test_ind : str
        The identifier for the test.
    sol_diff : NDArray[np.floating]
        The difference between the solution and the reference volume.
    tolerance : float
        The tolerance level for the deviation.
    """
    max_dev = np.max(np.abs(sol_diff))
    print(f"\n{test_ind} - Max absolute deviation is: {max_dev:.6} (tolerance: {tolerance:.6}) -> ", end="", flush=True)


def _fwd_op(M: NDArray, x: NDArray, y_shape: tuple[int, int]) -> NDArray:
    """
    Perform the forward operation using the projection matrix.

    Parameters
    ----------
    M : NDArray
        The projection matrix.
    x : NDArray
        The input volume.
    y_shape : tuple[int, int]
        The shape of the output data.

    Returns
    -------
    NDArray
        The projected data.
    """
    return np.dot(M, x.flatten()).reshape(y_shape)


# def _bwd_op(M: NDArray, y: NDArray, x_shape: tuple[int, int]) -> NDArray:
#     return np.dot(y.flatten(), M).reshape(x_shape)


@pytest.fixture(scope="module")
def rand_2d_data() -> tuple[NDArray, NDArray, NDArray]:
    """
    Generate random 2D data for testing.

    Returns
    -------
    tuple[NDArray, NDArray, NDArray]
        The projection matrix, the random volume, and the projected data.
    """
    vol_rand_2d = np.fmin(np.random.rand(*VOL_SHAPE_2D[:2]) + eps, 1)
    proj_matrix_2d = (np.random.rand(int(np.prod(PRJ_SHAPE_2D)), int(np.prod(VOL_SHAPE_2D))) > 0.5).astype(np.float32)
    data_rand_2d = _fwd_op(proj_matrix_2d, vol_rand_2d, tuple(PRJ_SHAPE_2D))
    return proj_matrix_2d, vol_rand_2d, data_rand_2d


@pytest.fixture(scope="module")
def flat_2d_data() -> tuple[NDArray, NDArray, NDArray]:
    """
    Generate flat 2D object data for testing.

    Returns
    -------
    tuple[NDArray, NDArray, NDArray]
        The projection matrix, the flat volume, and the projected data.
    """
    vol_flat_2d = processing.circular_mask(VOL_SHAPE_2D[:2], -2)
    proj_matrix_2d = (np.random.rand(int(np.prod(PRJ_SHAPE_2D)), int(np.prod(VOL_SHAPE_2D))) > 0.5).astype(np.float32)
    data_flat_2d = _fwd_op(proj_matrix_2d, vol_flat_2d, tuple(PRJ_SHAPE_2D))
    data_flat_2d += np.random.randn(*data_flat_2d.shape) * 1e-3
    return proj_matrix_2d, vol_flat_2d, data_flat_2d


@pytest.mark.parametrize("algo_it", [(solvers.SIRT, 1_000), (solvers.PDHG, 500), (solvers.MLEM, 10_000)])
def test_algo_rand(rand_2d_data, algo_it: tuple[type[solvers.Solver], int]):
    """
    Test algorithms in 2D on a random image.

    Parameters
    ----------
    rand_2d_data : tuple[NDArray, NDArray, NDArray]
        The random 2D data.
    algo_it : tuple[type[solvers.Solver], int]
        The algorithm class and the number of iterations.
    """
    proj_matrix_2d, vol_rand_2d, data_rand_2d = rand_2d_data
    algo_class, iterations = algo_it

    prj_mat = projectors.ProjectorMatrix(proj_matrix_2d, VOL_SHAPE_2D, PRJ_SHAPE_2D)

    algo = algo_class()
    sol, _ = algo(prj_mat, data_rand_2d, iterations=iterations)

    sol_diff = vol_rand_2d - sol
    _print_max_deviation(algo.info(), sol_diff, TOLERANCE)

    assert np.all(np.isclose(sol, vol_rand_2d, atol=TOLERANCE))


FLAT_2D_CASES = [
    (solvers.SIRT, 2_500, {}),
    (solvers.PDHG, 500, {}),
    (solvers.PDHG, 1_000, dict(precondition=False)),
    (solvers.PDHG, 500, dict(lower_limit=0, upper_limit=1)),
    (solvers.MLEM, 10_000, dict(lower_limit=0, upper_limit=1)),
]


@pytest.mark.parametrize("algo_it_pars", FLAT_2D_CASES)
def test_algo_flat(flat_2d_data, algo_it_pars: tuple[type[solvers.Solver], int, dict]):
    """
    Test algorithms in 2D on a flat 2D object.

    Parameters
    ----------
    flat_2d_object_data : tuple[NDArray, NDArray, NDArray]
        The flat 2D object data.
    algo_it_pars : tuple[type[solvers.Solver], int, dict]
        The algorithm class, the number of iterations, and extra parameters.
    """
    proj_matrix_2d, vol_flat_2d, data_flat_2d = flat_2d_data
    algo_class, iterations, extra_params = algo_it_pars

    prj_mat = projectors.ProjectorMatrix(proj_matrix_2d, VOL_SHAPE_2D, PRJ_SHAPE_2D)

    if algo_class in (solvers.SIRT, solvers.PDHG):
        reg = regularizers.Regularizer_TV2D(1e-4)
        algo = algo_class(regularizer=reg)
    else:
        algo = algo_class()
    sol, _ = algo(prj_mat, data_flat_2d, iterations=iterations, **extra_params)

    sol_diff = vol_flat_2d - sol
    _print_max_deviation(algo.info(), sol_diff, TOLERANCE)

    assert np.all(np.isclose(sol, vol_flat_2d, atol=TOLERANCE))



"""
Filtered back-projection filters.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Any, Optional, Union
import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import ArrayLike, DTypeLike, NDArray
from scipy.interpolate import interp1d
from skimage.transform.radon_transform import _get_fourier_filter

from .operators import BaseTransform
from .processing import circular_mask


try:
    import pywt

    has_pywt = True
except ImportError:
    print("WARNING: You need to install PyWavelets to benefit from wavelet bases.")

    has_pywt = False


class BasisOptions(ABC, Mapping):
    """Options for the different types of bases."""

    def __len__(self) -> int:
        """Return the number of options.

        Returns
        -------
        int
            The number of options.
        """
        return self.__dict__.__len__()

    def __getitem__(self, k: Any) -> Any:
        """Return the selected option.

        Parameters
        ----------
        k : Any
            The key of the selected option.

        Returns
        -------
        Any
            The selected option.
        """
        return self.__dict__.__getitem__(k)

    def __iter__(self) -> Any:
        """Iterate the options list.

        Returns
        -------
        Any
            The following option.
        """
        return iter(self.__dict__)


@dataclass
class BasisOptionsBlocks(BasisOptions):
    """Options for the wavelet bases."""

    binning_start: Optional[int] = 2
    binning_type: str = "exponential"
    order: int = 1
    normalized: bool = True


@dataclass
class BasisOptionsWavelets(BasisOptions):
    """Options for the wavelet bases."""

    wavelet: str = "bior2.2"
    level: int = 5
    norm: float = 1.0


def create_basis(
    num_pixels: int,
    binning_start: Optional[int] = 2,
    binning_type: str = "exponential",
    normalized: bool = False,
    order: int = 1,
    dtype: DTypeLike = np.float32,
) -> NDArray:
    """Compute filter basis matrix.

    Parameters
    ----------
    num_pixels : int
        Number of filter fixels.
    binning_start : Optional[int], optional
        Starting displacement of the binning, by default 2.
    binning_type : str, optional
        Type of pixel binning, by default "exponential".
    normalized : bool, optional
        Whether to normalize the bins by the window size, by default True.
    order : int, optional
        Order of the basis functions. Only 0 and 1 supported, by default 1.
    dtype : DTypeLike, optional
        Data type, by default np.float32.

    Returns
    -------
    NDArray
        The filter basis.
    """
    filter_positions = np.abs(np.fft.fftfreq(num_pixels, 1 / num_pixels))

    window_size = 1
    window_position = 0

    basis_r = []
    while window_position < filter_positions.max():
        basis_tmp = np.zeros(filter_positions.shape, dtype=dtype)

        if order == 0:
            binning_positions = np.logical_and(
                window_position <= filter_positions, filter_positions < (window_position + window_size)
            )

            basis_val = 1.0
            if normalized:
                basis_val /= window_size

            basis_tmp[binning_positions] = basis_val
        else:
            basis_tmp = np.fmax(1.0 - filter_positions / (window_position + window_size), 0.0)

        basis_r.append(basis_tmp)
        window_position += window_size

        if binning_start is not None and window_position > binning_start:
            if binning_type == "exponential":
                window_size = 2 * window_size
            elif binning_type == "incremental":
                window_size += 1
            elif binning_type == "custom":
                window_size += int(np.sqrt(window_position - binning_start))
            else:
                raise ValueError(f"Invalid 'binning_type' = {binning_type}.")

    basis_r = np.array(basis_r, dtype=dtype)

    if order > 0:
        for ii in range(1, basis_r.shape[0]):
            for jj in range(0, ii):
                other_vec = basis_r[jj, ...]
                other_vec_norm_2 = other_vec.dot(other_vec)
                basis_r[ii, ...] -= other_vec * other_vec.dot(basis_r[ii, ...]) / other_vec_norm_2
        if normalized:
            basis_r /= np.linalg.norm(basis_r, ord=1, axis=-1, keepdims=True)

    return basis_r


def create_basis_wavelet(
    num_pixels: int,
    wavelet: str = "bior2.2",
    level: int = 5,
    norm: float = 1.0,
    dtype: DTypeLike = np.float32,
) -> NDArray:
    """Compute filter basis matrix.
    Parameters
    ----------
    num_pixels : int
        Number of wavelet filters.
    wavelet: str, optional
        The wavelet to use, by default "bior4.4".
    level : int, optional
        The decomposition level to reach, by default 5.
    norm : float, optional
        The norm to use, by default 1.0.
    dtype : DTypeLike, optional
        Data type, by default np.float32.
    Returns
    -------
    NDArray
        The filter basis.
    """
    if not has_pywt:
        print("WARNING: You need to install PyWavelets to benefit from wavelet bases.")
        raise ImportError("PyWavelets (pywt) module not found.")

    # max_level = pywt.swt_max_level(num_pixels)
    # if level > max_level:
    #     print(f"WARNING: Requested level {level} is too high for {num_pixels} pixels. Max allowed is {max_level}.")
    #     level = max_level

    w = pywt.Wavelet(wavelet)

    dec_lo = np.trim_zeros(w.dec_lo)
    dec_hi = np.trim_zeros(w.dec_hi)

    crop_size_l = num_pixels // 2
    crop_size_u = num_pixels - crop_size_l

    pad_size_u = (num_pixels - len(dec_hi)) // 2
    pad_size_l = num_pixels - len(dec_hi) - pad_size_u
    pad_width = (pad_size_l, pad_size_u)

    basis_hi_tmp = np.pad(dec_hi, pad_width=np.array(pad_width))

    pad_size_u = (num_pixels - len(dec_lo)) // 2
    pad_size_l = num_pixels - len(dec_lo) - pad_size_u
    pad_width = (pad_size_l, pad_size_u)

    basis_lo_tmp = np.pad(dec_lo, pad_width=np.array(pad_width))

    coords = np.fft.fftshift(np.fft.fftfreq(num_pixels, 1 / num_pixels))

    basis_r = []
    basis_r.append(basis_hi_tmp)
    for _ in range(level - 1):
        basis_lo_old = basis_lo_tmp.copy()

        int_obj = interp1d(coords, basis_hi_tmp, kind="linear")
        basis_hi_tmp = int_obj((coords + 1 - len(coords) % 2) / 2)
        basis_hi_tmp = np.convolve(basis_hi_tmp, basis_lo_old, mode="same")

        int_obj = interp1d(coords, basis_lo_tmp, kind="linear")
        basis_lo_tmp = int_obj((coords + 1 - len(coords) % 2) / 2)
        basis_lo_tmp = np.convolve(basis_lo_tmp, basis_lo_old, mode="same")

        basis_r.append(basis_hi_tmp)
    basis_r.append(basis_lo_tmp)

    basis_r = np.array(basis_r, dtype=dtype)
    basis_r /= np.linalg.norm(basis_r, axis=-1, ord=norm, keepdims=True)

    return np.fft.ifftshift(basis_r, axes=(-1,))


class Filter(ABC):
    """Base FBP filter."""

    fbp_filter: NDArray[np.floating]
    pad_mode: str
    use_rfft: bool
    dtype: DTypeLike

    def __init__(
        self,
        fbp_filter: Union[ArrayLike, NDArray[np.floating], None],
        pad_mode: str,
        use_rfft: bool,
        dtype: DTypeLike,
    ) -> None:
        """Initialize Base FBP filter.

        Parameters
        ----------
        fbp_filter : Union[ArrayLike, NDArray[np.floating], None]
            The filter.
        pad_mode : str
            The padding mode.
        use_rfft : bool
            Whethert to use the `rfft` or complex `fft`.
        dtype : DTypeLike
            The data type of the filter.
        """
        self.dtype = dtype
        self.pad_mode = pad_mode.lower()
        self.use_rfft = use_rfft

        if fbp_filter is None:
            self.fbp_filter = np.array([1.0], dtype=dtype)
        else:
            self.fbp_filter = np.array(np.real(fbp_filter), dtype=dtype)

    def get_padding_size(self, data_wu_shape: Sequence[int]) -> int:
        """Compute the projection padding size for a linear convolution.

        Parameters
        ----------
        data_wu_shape : Sequence[int]
            The shape of the data

        Returns
        -------
        int
            The padding size of the last dimension.
        """
        return max(64, int(2 ** np.ceil(np.log2(2 * data_wu_shape[-1]))))

    def to_fourier(self, data_wu: NDArray) -> NDArray:
        if self.use_rfft:
            return np.fft.rfft(data_wu, axis=-1)
        else:
            return np.fft.fft(data_wu, axis=-1)

    def to_real(self, data_wu: NDArray) -> NDArray:
        if self.use_rfft:
            return np.fft.irfft(data_wu, axis=-1)
        else:
            return np.fft.ifft(data_wu, axis=-1).real

    @property
    def filter_fourier(self) -> NDArray[np.floating]:
        """Fourier representation of the filter.

        Returns
        -------
        NDArray[np.floating]
            The filter in Fourier.
        """
        return self.fbp_filter.copy()

    @property
    def filter_real(self) -> NDArray[np.floating]:
        """Real-space representation of the filter.

        Returns
        -------
        NDArray[np.floating]
            The filter in real-space.
        """
        fbp_filter_r = self.to_real(self.fbp_filter)
        return np.fft.fftshift(fbp_filter_r, axes=(-1,))

    @filter_real.setter
    def filter_real(self, fbp_filter_r: NDArray[np.floating]) -> None:
        self.fbp_filter = self.to_fourier(fbp_filter_r)

    @filter_fourier.setter
    def filter_fourier(self, fbp_filter_f: NDArray[np.floating]) -> None:
        self.fbp_filter = fbp_filter_f.copy()

    @property
    def num_filters(self) -> int:
        return np.array(self.fbp_filter, ndmin=2).shape[-2]

    def apply_filter(self, data_wu: NDArray, fbp_filter: Optional[NDArray] = None) -> NDArray:
        """Apply the filter to the data_wu.

        Parameters
        ----------
        data_wu : NDArray
            The sinogram.
        fbp_filter : NDArray, optional
            The filter to use. The default is None

        Returns
        -------
        NDArray
            The filtered sinogram.
        """
        data_wu_shape = data_wu.shape
        if fbp_filter is None:
            local_filter = self.fbp_filter
        else:
            local_filter = fbp_filter

        prj_size_pad = self.get_padding_size(data_wu_shape)

        pad_edge_u = (prj_size_pad - data_wu_shape[-1]) / 2
        pad_width = np.zeros((len(data_wu_shape), 2), dtype=int)
        pad_width[-1, :] = (int(np.ceil(pad_edge_u)), int(np.floor(pad_edge_u)))

        prj_pad = np.pad(data_wu, pad_width=tuple(pad_width), mode=self.pad_mode)  # type: ignore
        prj_pad = np.roll(prj_pad, shift=-pad_width[-1][0], axis=-1)

        prj_f = self.to_fourier(prj_pad)

        local_filter = np.array(local_filter, ndmin=2)

        num_filters = local_filter.shape[-2]

        if num_filters > 1:
            prjs_f = [np.array([])] * num_filters

            for ii, f in enumerate(local_filter):
                prj_f_ii = prj_f * np.array(f, ndmin=len(data_wu_shape))
                prj_f_ii = self.to_real(prj_f_ii)
                prjs_f[ii] = prj_f_ii[..., : data_wu_shape[-1]]

            return np.array(prjs_f)
        else:
            prj_f *= np.array(local_filter, ndmin=len(data_wu_shape))
            return self.to_real(prj_f)[..., : data_wu_shape[-1]]

    @abstractmethod
    def compute_filter(self, data_wu: NDArray) -> None:
        """Compute the FBP filter for the given data.

        Parameters
        ----------
        data_wu : NDArray
            The reference sinogram / projection data.
        """

    def __call__(self, data_wu: NDArray) -> NDArray:
        """Filter the sinogram, by first computing the filter, and then applying it.

        Parameters
        ----------
        data_wu : NDArray
            The unfiltered sinogram.

        Returns
        -------
        NDArray
            The filtered sinogram.
        """
        self.compute_filter(data_wu)
        return self.apply_filter(data_wu, self.fbp_filter)

    def plot_filters(self, fourier_abs: bool = False):
        filters_r = np.array(self.filter_real, ndmin=2)
        filters_f = np.array(self.filter_fourier, ndmin=2)

        f, axes = plt.subplots(1, 2, figsize=(10, 4))
        for ii in range(self.num_filters):
            axes[0].plot(filters_r[ii, ...], label=f"Filter-{ii}")
        axes[0].set_title("Real-space")
        axes[0].set_xlabel("Pixel")

        for ii in range(self.num_filters):
            filt_f = filters_f[ii, ...]
            if fourier_abs:
                filt_f = np.abs(filt_f)
            axes[1].plot(filt_f, label=f"Filter-{ii}")
        axes[1].set_title("Fourier-space")
        axes[1].set_xlabel("Frequency")

        axes[0].grid()
        axes[1].grid()
        if self.num_filters > 1:
            axes[1].legend()
        f.tight_layout()


class FilterCustom(Filter):
    """Custom FBP filter."""

    def __init__(
        self,
        fbp_filter: Union[ArrayLike, NDArray[np.floating], None],
        pad_mode: str = "constant",
        use_rfft: bool = True,
        dtype: DTypeLike = np.float32,
    ) -> None:
        """Initialize Custom FBP filter.

        Parameters
        ----------
        fbp_filter : Union[ArrayLike, NDArray[np.floating], None]
            The filter.
        pad_mode : str, optional
            The padding mode, by default "constant".
        use_rfft : bool, optional
            Whethert to use the `rfft` or complex `fft`, by default True.
        dtype : DTypeLike, optional
            The data type of the filter, by default np.float32.
        """
        super().__init__(fbp_filter, pad_mode, use_rfft, dtype)

    def compute_filter(self, data_wu: NDArray) -> None:
        """Provide dummy implementation, because it is not required."""


class FilterFBP(Filter):
    """Traditional FBP filter."""

    filter_name: str

    FILTERS = ("ramp", "shepp-logan", "cosine", "hamming", "hann")

    def __init__(
        self, filter_name: str = "ramp", pad_mode: str = "constant", use_rfft: bool = True, dtype: DTypeLike = np.float32
    ) -> None:
        """Initialize traditional FBP filter.

        Parameters
        ----------
        filter_name : str
            The filter name
        use_rfft : bool, optional
            Whethert to use the `rfft` or complex `fft`, by default True
        dtype : DTypeLike, optional
            The type of the filter, by default np.float32
        """
        if filter_name.lower() not in self.FILTERS:
            raise ValueError(f"Unknown filter {filter_name}. Available filters: {self.FILTERS}")

        super().__init__(fbp_filter=None, pad_mode=pad_mode, use_rfft=use_rfft, dtype=dtype)

        self.filter_name = filter_name.lower()

    def compute_filter(self, data_wu: NDArray) -> None:
        """Compute the traditional FBP filter for the given data.

        Parameters
        ----------
        data_wu : NDArray
            The reference sinogram / projection data.
        """
        prj_size_pad = self.get_padding_size(data_wu.shape)

        self.fbp_filter = _get_fourier_filter(prj_size_pad, self.filter_name)
        self.fbp_filter = np.squeeze(self.fbp_filter) * np.pi / (2 * data_wu.shape[-2])

        if self.use_rfft:
            self.fbp_filter = self.fbp_filter[: (self.fbp_filter.shape[-1]) // 2 + 1]

    def get_available_filters(self) -> Sequence[str]:
        """Provide available FBP filters.

        Returns
        -------
        Sequence[str]
            The available filters.
        """
        return self.FILTERS


class FilterMR(Filter):
    """Data dependent FBP filter.

    This is a simplified implementation from:

    [1] Pelt, D. M., & Batenburg, K. J. (2014). Improving filtered backprojection
    reconstruction by data-dependent filtering. Image Processing, IEEE
    Transactions on, 23(11), 4750-4762.

    Code inspired by: https://github.com/dmpelt/pymrfbp
    """

    projector: BaseTransform

    binning_type: str
    binning_start: Union[int, None]

    lambda_smooth: Union[float, None]

    is_initialized: bool

    def __init__(
        self,
        projector: BaseTransform,
        binning_type: str = "exponential",
        binning_start: Union[int, None] = 2,
        lambda_smooth: Optional[float] = None,
        pad_mode: str = "constant",
        use_rfft: bool = True,
        dtype: DTypeLike = np.float32,
    ) -> None:
        """Initialize data-driven FBP filter.

        Parameters
        ----------
        projector : BaseTransform
            The projector to use for handling the data.
        start_exp_binning : int, optional
            From which distance to start exponentional binning. The default is 2.
        lambda_smooth : float, optional
            Smoothing parameter. The default is None.
        dtype : DTypeLike, optional
            Filter data type. The default is np.float32.
        """
        super().__init__(fbp_filter=None, pad_mode=pad_mode, use_rfft=use_rfft, dtype=dtype)

        self.projector = projector

        self.binning_type = binning_type.lower()
        if self.binning_type not in ("exponential", "incremental"):
            raise ValueError("Binning type should be either 'exponential' or 'incremental'.")
        self.binning_start = binning_start
        self.lambda_smooth = lambda_smooth

        self.is_initialized = False

    def initialize(self, data_wu_shape: Sequence[int]) -> None:
        """Initialize filter.

        Parameters
        ----------
        data_wu_shape : Sequence[int]
            Shape of the data.
        """
        num_pad_pixels = self.get_padding_size(data_wu_shape)

        self.basis_r = create_basis(
            num_pad_pixels, binning_type=self.binning_type, binning_start=self.binning_start, dtype=self.dtype
        )

        self.basis_f = self.to_fourier(self.basis_r).real

        self.is_initialized = True

    def compute_filter(self, data_wu: NDArray) -> None:
        """Compute the filter.

        Parameters
        ----------
        data_wu : NDArray
            The sinogram.
        projector : ProjectorOperator
            The projector used in the FBP.
        """
        if not self.is_initialized:
            self.initialize(data_wu.shape)

        num_sino_pixels = data_wu.shape[-1]

        sino_size = data_wu.shape[-2] * num_sino_pixels
        nrows = sino_size
        ncols = self.basis_f.shape[-2]

        if self.lambda_smooth:
            grad_vol_size = num_sino_pixels * (num_sino_pixels - 1)
            nrows += 2 * grad_vol_size

        A = np.zeros((nrows, ncols), dtype=self.dtype)
        vol_mask = circular_mask([num_sino_pixels] * 2)

        for ii, bas_f in enumerate(self.basis_f):
            data_wu_f = self.apply_filter(data_wu, bas_f)

            img = self.projector.T(data_wu_f)
            img *= vol_mask

            A[:sino_size, ii] = self.projector(img).flatten()

            if self.lambda_smooth:
                dx = np.diff(img, axis=-2)
                dy = np.diff(img, axis=-1)
                d = np.concatenate((dx.flatten(), dy.flatten()))
                A[sino_size:, ii] = self.lambda_smooth * d

        b = np.zeros((nrows,), dtype=self.dtype)
        b[:sino_size] = data_wu.flatten()
        fitted_components = np.linalg.lstsq(A, b, rcond=None)[0].astype(self.dtype)

        self.fbp_filter = fitted_components.dot(self.basis_f)



# -*- coding: utf-8 -*-

"""Top-level package for PyCorrectedEmissionCT."""

from importlib import metadata

from . import models  # noqa: F401, F402

from . import operators  # noqa: F401, F402
from . import projectors  # noqa: F401, F402

from . import filters  # noqa: F401, F402
from . import data_terms  # noqa: F401, F402
from . import regularizers  # noqa: F401, F402
from . import solvers  # noqa: F401, F402

from . import param_tuning  # noqa: F401, F402
from . import processing  # noqa: F401, F402
from . import denoisers  # noqa: F401, F402

from . import struct_illum  # noqa: F401, F402
from . import alignment  # noqa: F401, F402

from . import physics  # noqa: F401, F402

from . import testing  # noqa: F401, F402


def get_version(dist: str = "corrct") -> str:
    """Get version of the given distribution.

    Parameters:
        dist: A distribution name.

    Returns:
        A version number.
    """
    try:
        return metadata.version(dist)
    except metadata.PackageNotFoundError:
        return "0.0.0"


__author__ = """Nicola VIGANÒ"""
__email__ = "N.R.Vigano@cwi.nl"
__version__ = get_version()



"""Units and conversion."""

from typing import overload, Union

import scipy.constants as spc
from numpy.typing import NDArray


class ConversionMetric:
    """Conversion factors between orders of magnitude of the metric units."""

    str_to_order = {
        "km": 1e-3,
        "m": 1e0,
        "cm": 1e2,
        "mm": 1e3,
        "um": 1e6,
        "nm": 1e9,
        "a": 1e10,
    }

    order_to_str = {
        1e-3: "km",
        1e0: "m",
        1e2: "cm",
        1e3: "mm",
        1e6: "um",
        1e9: "nm",
        1e10: "a",
    }

    @staticmethod
    def convert(from_unit: str, to_unit: str) -> float:
        """Convert numbers from the source unit to the destination unit.

        Parameters
        ----------
        from_unit : str
            The source unit
        to_unit : str
            The destination unit

        Returns
        -------
        float
            The conversion factor
        """
        return ConversionMetric.str_to_order[to_unit] / ConversionMetric.str_to_order[from_unit]


class ConversionEnergy:
    """Conversion factors between orders of magnitude of the energy units."""

    str_to_order = {
        "GeV": 1e-9,
        "MeV": 1e-6,
        "keV": 1e-3,
        "eV": 1e0,
        "meV": 1e3,
        "ueV": 1e6,
    }

    order_to_str = {
        1e-9: "GeV",
        1e-6: "MeV",
        1e-3: "keV",
        1e0: "eV",
        1e3: "meV",
        1e6: "ueV",
    }

    @staticmethod
    def convert(from_unit: str, to_unit: str) -> float:
        """Convert numbers from the source unit to the destination unit.

        Parameters
        ----------
        from_unit : str
            The source unit
        to_unit : str
            The destination unit

        Returns
        -------
        float
            The conversion factor
        """
        return ConversionEnergy.str_to_order[to_unit] / ConversionEnergy.str_to_order[from_unit]


@overload
def energy_to_wlength(energy: float, unit_wl: str = "m", unit_en: str = "keV") -> float: ...


@overload
def energy_to_wlength(energy: NDArray, unit_wl: str = "m", unit_en: str = "keV") -> NDArray: ...


def energy_to_wlength(energy: Union[float, NDArray], unit_wl: str = "m", unit_en: str = "keV") -> Union[float, NDArray]:
    """Convert from energy to wavelength.

    Parameters
    ----------
    energy : float | NDArray
        The energy
    unit_wl : str, optional
        The chosen unit for the output wavelength. The default is "m"
    unit_en : str, optional
        The chosen unit for the input energy. The default is "keV"

    Returns
    -------
    float | NDArray
        The wavelength in the chosen unit
    """
    factor_m = ConversionMetric.convert("m", unit_wl)
    factor_e = ConversionEnergy.convert("eV", unit_en)
    return factor_m * factor_e / (spc.physical_constants["electron volt-inverse meter relationship"][0] * energy)


@overload
def wlength_to_energy(w_length: float, unit_wl: str = "m", unit_en: str = "keV") -> float: ...


@overload
def wlength_to_energy(w_length: NDArray, unit_wl: str = "m", unit_en: str = "keV") -> NDArray: ...


def wlength_to_energy(w_length: Union[float, NDArray], unit_wl: str = "m", unit_en: str = "keV") -> Union[float, NDArray]:
    """Convert wavelength to energy.

    Parameters
    ----------
    w_length : float | NDArray
        The wavelength in the chosen unit
    unit : str, optional
        The chosen unit for the input wavelength. The default is "m"
    unit_en : str, optional
        The chosen unit for the output energy. The default is "keV"

    Returns
    -------
    float | NDArray
        The energy in the chosen unit
    """
    factor_m = ConversionMetric.convert("m", unit_wl)
    factor_e = ConversionEnergy.convert("eV", unit_en)
    return factor_m * factor_e / (spc.physical_constants["electron volt-inverse meter relationship"][0] * w_length)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test the examples.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import pytest

import os
import glob

import importlib

import matplotlib.pyplot as plt


examples_list = glob.glob(os.path.join("examples", "example_*"))


@pytest.mark.parametrize("example_file", examples_list)
def test_import(example_file: str):
    module_name = "examples." + os.path.split(example_file)[-1][:-3]
    print(f"\nTesting: {example_file}: {module_name}\n")
    importlib.import_module(module_name)
    plt.close("all")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Define all the models used through-out the code.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import numpy as np
from numpy.typing import ArrayLike, NDArray
from typing import Optional, Sequence, Union, Any

import scipy.spatial.transform as spt

from dataclasses import dataclass, replace as dc_replace

from abc import ABC

from copy import deepcopy


ROT_DIRS_VALID = ("clockwise", "counter-clockwise")


class Geometry(ABC):
    """Base geometry class."""

    def __str__(self) -> str:
        """
        Return a human readable representation of the object.

        Returns
        -------
        str
            The human readable representation of the object.
        """
        descr = f"{self.__class__.__name__}(\n"
        for f, v in self.__dict__.items():
            descr += f"    {f} = {v},\n"
        return descr + ")"


@dataclass
class ProjectionGeometry(Geometry):
    """Store the projection geometry."""

    geom_type: str
    src_pos_xyz: NDArray
    det_pos_xyz: NDArray
    det_u_xyz: NDArray
    det_v_xyz: NDArray
    rot_dir_xyz: NDArray
    pix2vox_ratio: float = 1
    det_shape_vu: Optional[NDArray] = None

    def __post_init__(self) -> None:
        self.geom_type = self.geom_type.lower()
        self.src_pos_xyz = np.array(self.src_pos_xyz, ndmin=2)
        self.det_pos_xyz = np.array(self.det_pos_xyz, ndmin=2)
        self.det_u_xyz = np.array(self.det_u_xyz, ndmin=2)
        self.det_v_xyz = np.array(self.det_v_xyz, ndmin=2)
        self.rot_dir_xyz = np.array(self.rot_dir_xyz, ndmin=2)

    def __getitem__(self, indx: Any):
        """
        Slice projection geometry along the angular direction.

        Parameters
        ----------
        indx : Any
            Indices of the slicing.
        """

        def slice_array(vecs_arr: NDArray, indx: Any):
            if len(vecs_arr.shape) > 1 and vecs_arr.shape[0] > 1:
                return vecs_arr[indx, :]
            else:
                return vecs_arr

        return dc_replace(
            self,
            src_pos_xyz=slice_array(self.src_pos_xyz, indx),
            det_pos_xyz=slice_array(self.det_pos_xyz, indx),
            det_u_xyz=slice_array(self.det_u_xyz, indx),
            det_v_xyz=slice_array(self.det_v_xyz, indx),
        )

    def copy(self) -> "ProjectionGeometry":
        """Deepcopy an existing geometry.

        Returns
        -------
        ProjectionGeometry
            The new instance of ProjectionGeometry
        """
        return deepcopy(self)

    @staticmethod
    def get_default_parallel(
        *,
        geom_type: str = "3d",
        rot_axis_shift_pix: Optional[ArrayLike] = None,
        rot_axis_dir: Union[str, ArrayLike] = "clockwise",
    ) -> "ProjectionGeometry":
        """
        Generate the default geometry for parallel beam.

        Parameters
        ----------
        geom_type : str, optional
            The geometry type. The default is "parallel3d".
        rot_axis_shift_pix : Optional[ArrayLike], optional
            Rotation axis shift in pixels. The default is None.
        rot_axis_dir : Union[str, ArrayLike], optional
            Rotation axis direction. It can be either a string or a direction. The default is "clockwise".

        Returns
        -------
        ProjectionGeometry
            The default parallel-beam geometry.
        """
        return get_prj_geom_parallel(geom_type=geom_type, rot_axis_shift_pix=rot_axis_shift_pix, rot_axis_dir=rot_axis_dir)

    @property
    def ndim(self) -> int:
        """Return the number of dimensions of the geometry.

        Returns
        -------
        int
            The number of dimensions.
        """
        if "parallel" in self.geom_type:
            return int(self.geom_type[-2])
        elif self.geom_type.lower() == "cone":
            return 3
        elif self.geom_type.lower() == "fanflat":
            return 2
        else:
            raise ValueError(
                f"Geometry ({self.geom_type}) needs to be one of: 'parallel2d' | 'parallel3d' | 'cone' | 'fanflat'."
            )

    def get_3d(self) -> "ProjectionGeometry":
        """Return the 3D version of the geometry.

        Returns
        -------
        ProjectionGeometry
            The new geometry.
        """
        if self.ndim == 2:
            if self.det_shape_vu is not None:
                new_det_shape_vu = np.ones(2, dtype=int)
                new_det_shape_vu[-len(self.det_shape_vu) :] = self.det_shape_vu
            else:
                new_det_shape_vu = None
            return dc_replace(self, geom_type=self.geom_type.replace("2d", "3d"), det_shape_vu=new_det_shape_vu)
        else:
            return dc_replace(self)

    def set_detector_shape_vu(self, vu: Union[int, Sequence[int], NDArray]) -> None:
        """Set the detector VU shape.

        Parameters
        ----------
        vu : int | Sequence[int] | NDArray
            The VU shape of the projection data.
        """
        self.det_shape_vu = np.array(vu, ndmin=1)

    def set_detector_shifts_vu(
        self,
        det_pos_vu: Union[ArrayLike, NDArray, None] = None,
        cor_pos_u: Union[float, None] = None,
        det_dist_y: ArrayLike = 0.0,
    ) -> None:
        """
        Set the detector position in XZ, from VU (vertical, horizontal) coordinates.

        Parameters
        ----------
        det_pos_vu : ArrayLike | NDArray | None
            Detector vertical and horizontal positions. Vertical is optional.
        cor_pos_u : float | None
            Center of rotation position along U.
        det_dist_y : ArrayLike, optional
            Detector distance from origin along Y. The default is 0.0.
        """
        det_pos_vu = np.array(det_pos_vu if det_pos_vu is not None else 0.0, ndmin=2, dtype=np.float64)
        if cor_pos_u is not None:
            det_pos_vu[-1, ...] = det_pos_vu[-1, ...] + cor_pos_u

        det_dist_y = np.array(det_dist_y, ndmin=1, dtype=np.float64)
        if det_dist_y.size > 1 and (
            det_dist_y.ndim > 1 or (det_pos_vu.shape[1] > 1 and det_dist_y.size != det_pos_vu.shape[1])
        ):
            raise ValueError(
                f"Detector distance along Y (shape: {det_dist_y.shape}) should either be a scalar or a 1D array of the "
                f"same length as the detector positions (shape: {det_pos_vu.shape}), if detector positions are more than 1."
            )

        if self.det_pos_xyz.shape[0] > 1 and det_pos_vu.shape[-1] > 1 and self.det_pos_xyz.shape[0] != det_pos_vu.shape[-1]:
            raise ValueError(
                f"Current number of angles ({self.det_pos_xyz.shape[-2]}) and new number of "
                f"angles ({det_pos_vu.shape[-1]}) differ!"
            )

        self.det_pos_xyz = np.zeros((det_pos_vu.shape[-1], 3), dtype=np.float64)
        self.det_pos_xyz += self.det_u_xyz * det_pos_vu[-1, :].reshape([-1, 1])
        if self.ndim == 3 and det_pos_vu.shape[0] == 2:
            self.det_pos_xyz += self.det_v_xyz * det_pos_vu[-2, :].reshape([-1, 1])

        self.det_pos_xyz[:, 1] += det_dist_y

    def set_source_shifts_vu(self, src_pos_vu: Union[ArrayLike, NDArray, None] = None) -> None:
        """
        Set the source position in XZ, from VU (vertical, horizontal) coordinates.

        Parameters
        ----------
        src_pos_vu : ArrayLike | NDArray | None
            Source vertical and horizontal positions. Vertical is optional.
        """
        if src_pos_vu is None:
            return

        src_pos_vu = np.array(src_pos_vu, ndmin=2)

        if self.src_pos_xyz.shape[0] > 1 and src_pos_vu.shape[-1] > 1 and self.src_pos_xyz.shape[0] != src_pos_vu.shape[-1]:
            raise ValueError(
                f"Current number of angles ({self.src_pos_xyz.shape[-2]}) and new number of angles ({src_pos_vu.shape[-1]}) differ!"
            )
        src_pos_y = self.src_pos_xyz[:, 1].copy()
        self.src_pos_xyz = np.zeros((src_pos_vu.shape[-1], 3))
        self.src_pos_xyz[:, 0] = src_pos_vu[-1, :]
        self.src_pos_xyz[:, 1] = src_pos_y
        if self.ndim == 3 and src_pos_vu.shape[0] == 2:
            self.src_pos_xyz[:, 2] = src_pos_vu[-2, :]

    def set_detector_tilt(
        self,
        angles_t_rad: Union[ArrayLike, NDArray],
        tilt_axis: Union[Sequence[float], NDArray] = (0, 1, 0),
        tilt_source: bool = False,
    ) -> None:
        """
        Rotate the detector by the given angle(s) and axis(axes).

        Parameters
        ----------
        angles_t_rad : ArrayLike | NDArray
            Rotation angle(s) in radians.
        tilt_axis : Sequence[float] | NDArray, optional
            The tilt axis or axes. The default is (0, 1, 0)
        tilt_source : bool, optional
            Whether to also tilt the source. The default is False.

        Notes
        -----
        When applying multiple axes, they will be applied in order. This means
        that the application is not going to be independent.
        """
        angles = np.array(angles_t_rad, ndmin=1)[:, None]
        tilt_axis = np.array(tilt_axis, ndmin=1)
        if tilt_axis.shape[-1] != 3:
            raise ValueError(
                f"Tilt axis/axes should be three-dimensional, along the last dimension. Current shape: {tilt_axis.shape}"
            )
        if tilt_axis.ndim == 1:
            tilt_axis = tilt_axis[None, :]
        elif tilt_axis.ndim > 2:
            raise ValueError(
                f"Tilt axis/axes should be three-dimensional, along the last dimension. Current shape: {tilt_axis.shape}"
            )
        elif angles.size > 1 and tilt_axis.shape[0] != angles.shape[0]:
            raise ValueError(
                "Tilt axes and tilt angles multiplicity should match. "
                f"Current shapes: {tilt_axis.shape = }, {angles.shape = }"
            )

        for angle, axis in zip(angles, tilt_axis):
            rotations = spt.Rotation.from_rotvec(angle * axis)  # type: ignore

            if tilt_source:
                self.src_pos_xyz = rotations.apply(self.src_pos_xyz)

            self.det_u_xyz = rotations.apply(self.det_u_xyz)
            self.det_v_xyz = rotations.apply(self.det_v_xyz)

            self.det_pos_xyz = rotations.apply(self.det_pos_xyz)

    def rotate(self, angles_w_rad: ArrayLike, patch_astra_2d: bool = False) -> "ProjectionGeometry":
        """
        Rotate the geometry by the given angle(s).

        Parameters
        ----------
        angles_w_rad : ArrayLike
            Rotation angle(s) in radians.

        Returns
        -------
        ProjectionGeometry
            The rotated geometry.
        """
        angles = np.array(angles_w_rad, ndmin=1)[:, None]

        # Dealing with ASTRA's incoherent 2D and 3D coordinate systems.
        if patch_astra_2d and self.ndim == 2:
            angles = -angles

        rotations = spt.Rotation.from_rotvec(angles * self.rot_dir_xyz)  # type: ignore

        return dc_replace(
            self,
            src_pos_xyz=rotations.apply(self.src_pos_xyz),
            det_pos_xyz=rotations.apply(self.det_pos_xyz),
            det_u_xyz=rotations.apply(self.det_u_xyz),
            det_v_xyz=rotations.apply(self.det_v_xyz),
        )

    def get_field_scaled(self, field_name: str) -> NDArray:
        """
        Return the a field content, scaled by the pix2vox ratio.

        Parameters
        ----------
        field_name : str
            Name of the field to access.

        Returns
        -------
        NDArray
            The scaled field.
        """
        field_value = getattr(self, field_name) / self.pix2vox_ratio
        if self.geom_type.lower() != "cone" and int(self.geom_type[-2]) == 2:
            return field_value[:, :-1]
        else:
            return field_value

    def project_displacement_to_detector(self, disp_zyx: ArrayLike) -> NDArray:
        """Project a given displacement vector in the volume coordinates, over the detector.

        Parameters
        ----------
        disp_zyx : ArrayLike
            The displacement vector in volume coordinates.

        Returns
        -------
        NDArray
            The projection on u (and if applicable v) coordinates.

        Raises
        ------
        ValueError
            When projection geometry and vector dimensions don match.
        """
        disp_zyx = np.array(disp_zyx, ndmin=1)

        disp_dims = len(disp_zyx)
        if self.ndim != disp_dims:
            raise ValueError(f"Geometry is {self.ndim}d, while passed displacement is {disp_dims}d.")

        disp_xyz = np.flip(disp_zyx)

        if self.ndim == 2:
            return self.det_u_xyz[..., : self.ndim].dot(disp_xyz)
        else:
            return np.stack(
                [self.det_v_xyz[..., : self.ndim].dot(disp_xyz), self.det_u_xyz[..., : self.ndim].dot(disp_xyz)], axis=0
            )

    def get_pre_weights(self, det_shape_vu: Union[Sequence[int], NDArray, None] = None) -> Union[NDArray, None]:
        """Compute the pre-weights of the projector geometry (notably for cone-beam geometries).

        Parameters
        ----------
        det_shape_vu : Sequence[int] | NDArray | None, optional
            Shape of the detector in [V]U coordinates, by default None

        Returns
        -------
        NDArray | None
            The computed detector weights
        """
        if self.geom_type != "cone":
            return None
        else:
            if det_shape_vu is None:
                if self.det_shape_vu is None:
                    print("WARNING: pre-weights cannot be computed because detector shape is None.")
                    return None
                else:
                    det_shape_vu = self.det_shape_vu
            det_shape_vu = np.array(det_shape_vu, dtype=int)
            if self.det_shape_vu is not None and np.any(det_shape_vu != self.det_shape_vu):
                print("WARNING: overriding the detector shape in the computation of the pre-weights.")

            src2det_xyz = self.det_pos_xyz + self.src_pos_xyz

            pixel_coords_vu = [np.linspace(-s / 2, s / 2, int(s)) for s in det_shape_vu]
            pixel_coords_vu = np.meshgrid(*pixel_coords_vu, indexing="ij")
            pixel_coords_vu = [coords[..., None, None] for coords in pixel_coords_vu]

            pixel_coords_xyz = pixel_coords_vu[-1] * self.det_u_xyz
            if len(pixel_coords_vu) > 1:
                pixel_coords_xyz += pixel_coords_vu[-2] * self.det_v_xyz

            src2pixel_dict = np.linalg.norm(pixel_coords_xyz + src2det_xyz, axis=-1)
            src2det_dist = np.linalg.norm(src2det_xyz, axis=-1)

            pre_weights = src2det_dist / (src2pixel_dict + (src2pixel_dict == 0))
            return pre_weights.swapaxes(-2, -1)


@dataclass
class VolumeGeometry(Geometry):
    """Store the volume geometry."""

    _vol_shape_xyz: NDArray
    vox_size: float = 1.0

    def __post_init__(self):
        """Initialize the input parameters."""
        self._vol_shape_xyz = np.array(self._vol_shape_xyz, ndmin=1)

    def is_square(self) -> bool:
        """Compute whether the volume is square in XY.

        Returns
        -------
        bool
            True is the volume is square in XY.
        """
        return self._vol_shape_xyz[0] == self._vol_shape_xyz[1]

    @property
    def shape_xyz(self) -> NDArray:
        """
        Return the volume shape (XYZ).

        Returns
        -------
        NDArray
            Shape of the volume (XYZ).
        """
        return self._vol_shape_xyz

    @property
    def shape_zxy(self) -> NDArray:
        """
        Return the volume shape (ZXY).

        The swap between X and Y is imposed by the astra-toolbox.

        Returns
        -------
        NDArray
            Shape of the volume (ZXY).
        """
        vol_shape_zyx = np.flip(self._vol_shape_xyz)
        return np.array([*vol_shape_zyx[:-2], vol_shape_zyx[-1], vol_shape_zyx[-2]], dtype=int)

    @property
    def mask_shape(self) -> NDArray:
        """Return the XY volume shape for circular masks.

        Returns
        -------
        NDArray
            Shape of the XY volume.
        """
        return self.shape_xyz[:2]

    @property
    def extent(self) -> Sequence[float]:
        """
        Return extent of the volume.

        Returns
        -------
        Sequence[float]
            The extent of the volume [-x, +x, -y, +y, [-z, +z]].
        """
        half_size_xyz = self._vol_shape_xyz * self.vox_size / 2
        return [hs * sign for hs in half_size_xyz for sign in [-1, +1]]

    def is_3D(self) -> bool:
        """
        Tell whether this is a 3D geometry.

        Returns
        -------
        bool
            Whether this is a 3D geometry or not.
        """
        return len(self._vol_shape_xyz) == 3 and self._vol_shape_xyz[-1] > 1

    def get_3d(self) -> "VolumeGeometry":
        """Return the 3D version of the geometry.

        Returns
        -------
        VolumeGeometry
            The new geometry.
        """
        if len(self._vol_shape_xyz) == 2:
            return dc_replace(self, _vol_shape_xyz=np.concatenate((self._vol_shape_xyz, [1])))
        else:
            return dc_replace(self)

    @staticmethod
    def get_default_from_data(data: NDArray, data_format: str = "dvwu") -> "VolumeGeometry":
        """
        Generate a default volume geometry from the data shape.

        Parameters
        ----------
        data : NDArray
            The data.
        data_format : str, optional
            The ordering and meaning of the dimensions in the data. The deault is "dvwu".

        Returns
        -------
        VolumeGeometry
            The default volume geometry.
        """
        return get_vol_geom_from_data(data=data, data_format=data_format)

    @staticmethod
    def get_default_from_volume(volume: NDArray) -> "VolumeGeometry":
        """
        Generate a default volume geometry from the given volume.

        Parameters
        ----------
        volume : NDArray
            The volume.

        Returns
        -------
        VolumeGeometry
            The default volume geometry.
        """
        return get_vol_geom_from_volume(volume=volume)


def combine_shifts_vu(shifts_v: NDArray, shifts_u: NDArray) -> NDArray:
    """Combine vertical and horizontal shifts.

    Parameters
    ----------
    shifts_v : NDArray
        The vertical shifts
    shifts_u : NDArray
        The horizontal shifts

    Returns
    -------
    NDArray
        The combined shifts
    """
    if np.sum(np.array(shifts_v.shape) > 1) > 1:
        raise ValueError(f"Expected 1-dimensional array for vertical shifts, but an array {shifts_v.shape = } was passed")
    if np.sum(np.array(shifts_u.shape) > 1) > 1:
        raise ValueError(f"Expected 1-dimensional array for horizontal shifts, but an array {shifts_u.shape = } was passed")
    if shifts_v.size != shifts_u.size:
        raise ValueError(f"Number of vertical shifts ({shifts_v.size}) and horizontal shifts ({shifts_u.size}) should match")

    return np.stack([np.squeeze(shifts_v), np.squeeze(shifts_u)], axis=-2)


def get_rot_axis_dir(rot_axis_dir: Union[str, ArrayLike, NDArray] = "clockwise") -> NDArray:
    """Process the requested rotation axis direction and return a meaningful value.

    Parameters
    ----------
    rot_axis_dir : Union[str, ArrayLike, NDArray], optional
        The requested direction, by default "clockwise"

    Returns
    -------
    NDArray
        The vector corresponding to the rotation direction.

    Raises
    ------
    ValueError
        In case of malformed direction.
    """
    if isinstance(rot_axis_dir, str):
        if rot_axis_dir.lower() not in ROT_DIRS_VALID:
            raise ValueError(f"Rotation axis direction {rot_axis_dir} not allowed. It should be one of: {ROT_DIRS_VALID}")

        if rot_axis_dir.lower() == "clockwise":
            return np.array([0.0, 0.0, -1.0])
        else:
            return np.array([0.0, 0.0, 1.0])
    else:
        return np.array(rot_axis_dir, ndmin=1)


def _get_data_dims(data_shape: Union[Sequence[int], NDArray], data_format: str = "dvwu") -> dict[str, Union[int, None]]:
    dims: dict[str, Union[int, None]] = dict(u=None, v=None, w=None, d=None)
    for ii in range(-len(data_shape), 0):
        dims[data_format[ii]] = data_shape[ii]
    return dims


def get_prj_geom_parallel(
    *,
    geom_type: str = "3d",
    rot_axis_shift_pix: Union[ArrayLike, NDArray, None] = None,
    rot_axis_dir: Union[str, ArrayLike, NDArray] = "clockwise",
    data_shape: Union[Sequence[int], NDArray, None] = None,
    data_format: str = "dvwu",
) -> ProjectionGeometry:
    """
    Generate the default geometry for parallel beam.

    Parameters
    ----------
    geom_type : str, optional
        The geometry type. The default is "parallel3d".
    rot_axis_shift_pix : ArrayLike | NDArray | None, optional
        Rotation axis shift in pixels. The default is None.
    rot_axis_dir : str | ArrayLike | NDArray, optional
        Rotation axis direction. It can be either a string or a direction. The default is "clockwise".

    Returns
    -------
    ProjectionGeometry
        The default parallel-beam geometry.
    """
    geom_type = geom_type.lower()

    prj_geom = ProjectionGeometry(
        geom_type="parallel" + geom_type,
        src_pos_xyz=np.array([0.0, -1.0, 0.0]),
        det_pos_xyz=np.zeros(3),
        det_u_xyz=np.array([1.0, 0.0, 0.0]),
        det_v_xyz=np.array([0.0, 0.0, 1.0]),
        rot_dir_xyz=get_rot_axis_dir(rot_axis_dir),
    )

    if rot_axis_shift_pix is not None:
        rot_axis_shift_pix = np.array(rot_axis_shift_pix)
        if rot_axis_shift_pix.size == 1:
            prj_geom.set_detector_shifts_vu(cor_pos_u=float(rot_axis_shift_pix))
        else:
            prj_geom.set_detector_shifts_vu(det_pos_vu=rot_axis_shift_pix)

    if data_shape is not None:
        data_dims = _get_data_dims(data_shape, data_format)
        if data_dims["u"] is None:
            raise ValueError(
                "Could not determine data dimensions. Coordinate U cannot be undetermined."
                f" Data shape: {data_shape}, data format: {data_format}"
            )
        if geom_type == "3d":
            if data_dims["v"] is None:
                raise ValueError(
                    "Could not determine data dimensions. Coordinate V cannot be undetermined in a 3D geometry."
                    f" Data shape: {data_shape}, data format: {data_format}"
                )
            prj_geom.set_detector_shape_vu([data_dims["v"], data_dims["u"]])
        else:
            prj_geom.set_detector_shape_vu([data_dims["u"]])

    return prj_geom


def get_prj_geom_cone(
    *,
    src_to_sam_dist: float,
    rot_axis_shift_pix: Union[ArrayLike, NDArray, None] = None,
    rot_axis_dir: Union[str, ArrayLike, NDArray] = "clockwise",
    data_shape: Union[Sequence[int], NDArray, None] = None,
    data_format: str = "dvwu",
) -> ProjectionGeometry:
    """
    Generate the default geometry for parallel beam.

    Parameters
    ----------
    geom_type : str, optional
        The geometry type. The default is "parallel3d".
    rot_axis_shift_pix : ArrayLike | NDArray | None, optional
        Rotation axis shift in pixels. The default is None.
    rot_axis_dir : str | ArrayLike | NDArray, optional
        Rotation axis direction. It can be either a string or a direction. The default is "clockwise".

    Returns
    -------
    ProjectionGeometry
        The default cone-beam geometry.
    """
    prj_geom = ProjectionGeometry(
        geom_type="cone",
        src_pos_xyz=np.array([0.0, -src_to_sam_dist, 0.0]),
        det_pos_xyz=np.zeros(3),
        det_u_xyz=np.array([1.0, 0.0, 0.0]),
        det_v_xyz=np.array([0.0, 0.0, 1.0]),
        rot_dir_xyz=get_rot_axis_dir(rot_axis_dir),
    )

    if rot_axis_shift_pix is not None:
        rot_axis_shift_pix = np.array(rot_axis_shift_pix)
        if rot_axis_shift_pix.size == 1:
            prj_geom.set_detector_shifts_vu(cor_pos_u=float(rot_axis_shift_pix))
        else:
            prj_geom.set_detector_shifts_vu(det_pos_vu=rot_axis_shift_pix)

    if data_shape is not None:
        data_dims = _get_data_dims(data_shape, data_format)
        if data_dims["v"] is None or data_dims["u"] is None:
            raise ValueError(
                "Could not determine data dimensions. Coordinates UV cannot be undetermined in a cone-beam geometry."
                f" Data shape: {data_shape}, data format: {data_format}"
            )
        prj_geom.set_detector_shape_vu([data_dims["v"], data_dims["u"]])

    return prj_geom


def get_vol_geom_from_data(
    data: NDArray, padding_u: Union[int, Sequence[int], NDArray] = 0, data_format: str = "dvwu", super_sampling: int = 1
) -> VolumeGeometry:
    """
    Generate a default volume geometry from the data shape.

    Parameters
    ----------
    data : NDArray
        The data.
    padding_u : int | Sequence[int]
    data_format : str, optional
        The ordering and meaning of the dimensions in the data. The default is "dvwu".
    super_sampling: int, optional
        The super-sampling size of the voxels. The default is 1.

    Returns
    -------
    VolumeGeometry
        The default volume geometry.
    """
    data_dims = _get_data_dims(data.shape, data_format)
    if data_dims["u"] is None:
        raise ValueError(
            "Could not determine data dimensions. Coordinate U cannot be undetermined."
            f" Data shape: {data.shape}, data format: {data_format}"
        )

    if isinstance(padding_u, (Sequence, np.ndarray)):
        if len(padding_u) != 2:
            raise ValueError(
                f"Padding along U can only either be an integer or a Sequence/NDArray of 2 values. {padding_u} passed instead."
            )
        data_dims["u"] -= padding_u[0] + padding_u[1]
    else:
        data_dims["u"] -= padding_u * 2

    dims_xyz = [data_dims["u"]] * 2
    if data_dims["v"] is not None:
        dims_xyz.append(data_dims["v"])

    return VolumeGeometry(np.array(dims_xyz) * super_sampling, vox_size=1 / super_sampling)


def get_vol_geom_from_volume(volume: NDArray) -> VolumeGeometry:
    """
    Generate a default volume geometry from the given volume.

    Parameters
    ----------
    volume : NDArray
        The volume.

    Returns
    -------
    VolumeGeometry
        The default volume geometry.
    """
    vol_shape_zxy = volume.shape
    if len(vol_shape_zxy) < 2:
        raise ValueError(f"The volume should be at least 2-dimensional, but the following shape was passed: {vol_shape_zxy}")
    return VolumeGeometry(np.array([vol_shape_zxy[-2], vol_shape_zxy[-1], *np.flip(vol_shape_zxy[:-2])]))



"""
Tomographic projectors.

@author: Nicola VIGANÒ, Computational Imaging group, CWI, The Netherlands,
and ESRF - The European Synchrotron, Grenoble, France
"""

import concurrent.futures as cf
import multiprocessing as mp
from collections.abc import Sequence
from typing import Optional
from typing import Union
import numpy as np
from numpy.typing import ArrayLike
from numpy.typing import DTypeLike
from numpy.typing import NDArray
from scipy.sparse import spmatrix
from corrct import _projector_backends as prj_backends
from corrct import models
from corrct import operators
from corrct.physics.attenuation import AttenuationVolume


num_threads = round(np.log2(mp.cpu_count() + 1))
astra_available = prj_backends.has_astra and prj_backends.has_cuda


class ProjectorMatrix(operators.ProjectorOperator):
    """Projector that uses an explicit projection matrix."""

    A: Union[NDArray, spmatrix]

    def __init__(self, A: Union[NDArray, spmatrix], vol_shape: ArrayLike, prj_shape: ArrayLike) -> None:
        """
        Initialize the matrix projector.

        Parameters
        ----------
        A : NDArray | spmatrix
            The projection matrix.
        vol_shape : ArrayLike
            Volume shape.
        prj_shape : ArrayLike
            Projection shape.
        """
        self.vol_shape = np.array(vol_shape, ndmin=1, dtype=int)
        self.prj_shape = np.array(prj_shape, ndmin=1, dtype=int)

        self.A = A
        super().__init__()

    def _transpose(self) -> operators.ProjectorOperator:
        """
        Create the transpose operator.

        Returns
        -------
        operators.ProjectorOperator
            The transpose operator.
        """
        return ProjectorMatrix(self.A.transpose(), self.prj_shape, self.vol_shape)

    def absolute(self) -> operators.ProjectorOperator:
        """
        Return the projection operator using the absolute value of the projection coefficients.

        Returns
        -------
        operators.ProjectorOperator
            The absolute value operator.
        """
        return ProjectorMatrix(np.abs(self.A), self.vol_shape, self.prj_shape)  # type: ignore

    def fp(self, x: NDArray) -> NDArray:
        """
        Define the interface for the forward-projection.

        Parameters
        ----------
        x : NDArray
            Input volume.

        Returns
        -------
        NDArray
            The projection data.
        """
        return self.A.dot(x.flatten()).reshape(self.prj_shape)

    def bp(self, x: NDArray) -> NDArray:
        """
        Define the interface for the back-projection.

        Parameters
        ----------
        x : NDArray
            Input projection data.

        Returns
        -------
        NDArray
            The back-projected volume.
        """
        return self.A.transpose().dot(x.flatten()).reshape(self.vol_shape)


class ProjectorUncorrected(operators.ProjectorOperator):
    """Base projection class."""

    vol_geom: models.VolumeGeometry
    projector_backend: prj_backends.ProjectorBackend

    prj_intensities: Union[NDArray[np.floating], None]
    psf: Union[NDArray[np.floating], float, None]

    def __init__(
        self,
        vol_geom: Union[Sequence[int], NDArray[np.integer], models.VolumeGeometry],
        angles_rot_rad: Union[Sequence[float], NDArray],
        rot_axis_shift_pix: Union[float, ArrayLike, NDArray, None] = None,
        *,
        prj_geom: Optional[models.ProjectionGeometry] = None,
        prj_intensities: Optional[ArrayLike] = None,
        psf: Optional[ArrayLike] = None,
        backend: Union[str, prj_backends.ProjectorBackend] = "astra" if astra_available else "skimage",
        create_single_projs: bool = False,
    ):
        """Initialize the base projection class.

        It implements the forward and back projection of the single lines of a sinogram.
        It takes care of initializing and disposing the ASTRA projectors when used in a *with* statement.
        It supports both 2D and 3D geometries.

        Parameters
        ----------
        vol_geom : Sequence[int] | NDArray[np.integer] | models.VolumeGeometry
            The volume shape in Y X and optionally Z.
        angles_rot_rad : Sequence[float] | NDArray
            The rotation angles.
        rot_axis_shift_pix : float | ArrayLike | NDArray, optional
            The rotation axis shift(s) in pixels, by default None.
        prj_geom : models.ProjectionGeometry | None, optional
            The fully specified projection geometry.
            When active, the rotation axis shift is ignored, by default None.
        prj_intensities : ArrayLike | NDArray | None, optional
            Projection scaling factor, by default None.
        psf : ArrayLike | NDArray | None, optional
            The "point spread function" of the detector, by default None.
        backend : bool, optional
            Whether to use ASTRA or fall back to scikit-image.
            The default is True if CUDA and ASTRA are available, otherwise False.
        create_single_projs : bool, optional
            Whether to create projectors for single projections.
            Used for corrections and SART, by default False.

        Raises
        ------
        ValueError
            When the geometry is not correct.
        """
        if not astra_available:
            astra_status = f"astra: {prj_backends.has_astra}, cuda: {prj_backends.has_cuda}"
            if isinstance(backend, str) and backend == "astra":
                backend = "skimage"
                print(f"WARNING: ASTRA backend requested but not available ({astra_status}). Falling back to scikit-image.")
            elif isinstance(backend, prj_backends.ProjectorBackendASTRA):
                raise ValueError(f"Passed ASTRA projector, but astra not available ({astra_status}).")

        if not isinstance(vol_geom, models.VolumeGeometry):
            vol_geom = models.VolumeGeometry(_vol_shape_xyz=np.array(vol_geom))
        self.vol_geom = vol_geom
        self.prj_geom = prj_geom

        if not len(self.vol_geom.shape_xyz) in (2, 3):
            raise ValueError("Only 2D or 3D volumes are valid")
        if not self.vol_geom.is_square():
            raise ValueError("Only square volumes are valid")

        if isinstance(backend, str):
            if backend == "astra":
                if prj_backends.has_astra_direct:
                    self.projector_backend = prj_backends.ProjectorBackendDirectASTRA()
                else:
                    self.projector_backend = prj_backends.ProjectorBackendASTRA()
            elif backend == "skimage":
                self.projector_backend = prj_backends.ProjectorBackendSKimage()
            else:
                raise ValueError(f"Unknown backend: {backend}. Available options are: 'astra', 'skimage'.")
        else:
            self.projector_backend = backend

        if not (prj_geom is None or isinstance(self.projector_backend, prj_backends.ProjectorBackendASTRA)):
            raise ValueError("Using class `ProjectionGeometry` requires using astra-toolbox.")

        self.projector_backend.initialize_geometry(
            vol_geom=vol_geom,
            angles_rot_rad=angles_rot_rad,
            rot_axis_shift_pix=rot_axis_shift_pix,
            prj_geom=prj_geom,
            create_single_projs=create_single_projs,
        )

        if prj_intensities is not None:
            prj_intensities = np.array(prj_intensities, dtype=np.floating)
        self.prj_intensities = prj_intensities

        self._set_psf(psf)

        self.vol_shape = np.array(self.projector_backend.get_vol_shape(), ndmin=1)
        self.prj_shape = np.array(self.projector_backend.get_prj_shape(), ndmin=1)
        super().__init__()

    @property
    def angles_rot_rad(self) -> NDArray:
        """Simplify access to the rotation angles (in radians).

        Returns
        -------
        NDArray
            The rotation angles (in radians).
        """
        return self.projector_backend.angles_w_rad

    def __enter__(self):
        """Initialize the with statement block."""
        self.projector_backend.make_ready()
        return self

    def __exit__(self, *args):
        """De-initialize the with statement block."""
        self.projector_backend.dispose()

    def _set_psf(self, psf: Optional[ArrayLike], is_conv_symm: bool = False) -> None:
        if psf is not None:
            psf = np.squeeze(np.array(psf))
            if len(psf.shape) >= len(self.vol_geom.shape_xyz):
                raise ValueError(
                    "PSF should either be 1D (for 2D and 3D reconstructions) or 2D (for 3D reconstructions)."
                    + f" Passed PSF has shape: {psf.shape}, and the reconstruction is {len(self.vol_geom.shape_xyz)}D."
                )
            # This redundancy is required, due to the different results between the single-angle and multi-angle projections
            prj_shape_vu = [*self.projector_backend.prj_shape_vu[:-2], self.projector_backend.prj_shape_vu[-1]]
            prj_shape_vu = np.array(prj_shape_vu, ndmin=1)
            self.psf_vu = operators.TransformConvolution(prj_shape_vu, kernel=psf, is_symm=is_conv_symm)
            prj_shape_vwu = np.array(self.projector_backend.prj_shape_vwu, ndmin=1)
            self.psf_vwu = operators.TransformConvolution(prj_shape_vwu, kernel=psf[..., None, :], is_symm=is_conv_symm)
        else:
            self.psf_vu = self.psf_vwu = None

    def get_pre_weights(self) -> Union[NDArray, None]:
        """Compute the pre-weights of the projector geometry (notably for cone-beam geometries).

        Returns
        -------
        Union[NDArray, None]
            The computed detector weights
        """
        if self.prj_geom is None:
            return None
        else:
            return self.prj_geom.get_pre_weights([*self.prj_shape[-3:-2], self.prj_shape[-1]])

    def fp_angle(self, vol: NDArray, angle_ind: int) -> NDArray:
        """Forward-project a volume to a single sinogram line.

        Parameters
        ----------
        vol : NDArray
            The volume to forward-project.
        angle_ind : int
            The angle index to forward project.

        Returns
        -------
        x : NDArray
            The forward-projected sinogram line.
        """
        prj_vu = self.projector_backend.fp(vol, angle_ind)
        if self.prj_intensities is not None:
            prj_vu *= self.prj_intensities[angle_ind]
        if self.psf_vu is not None:
            prj_vu = self.psf_vu(prj_vu)
        return prj_vu

    def bp_angle(self, prj_vu: NDArray, angle_ind: int) -> NDArray:
        """Back-project a single sinogram line to the volume.

        Parameters
        ----------
        prj_vu : NDArray
            The sinogram to back-project or a single line.
        angle_ind : int
            The angle index to forward project.

        Returns
        -------
        NDArray
            The back-projected volume.
        """
        if self.prj_intensities is not None:
            prj_vu = prj_vu * self.prj_intensities[angle_ind]  # It will make a copy
        if self.psf_vu is not None:
            prj_vu = self.psf_vu.T(prj_vu)
        return self.projector_backend.bp(prj_vu, angle_ind)

    def fp(self, vol: NDArray) -> NDArray:
        """
        Forward-projection of the volume to the projection data.

        Parameters
        ----------
        vol : NDArray
            The volume to forward-project.

        Returns
        -------
        NDArray
            The forward-projected projection data.
        """
        prj_vwu = self.projector_backend.fp(vol)
        if self.prj_intensities is not None:
            prj_vwu *= self.prj_intensities[:, np.newaxis]
        if self.psf_vwu is not None:
            prj_vwu = self.psf_vwu(prj_vwu)
        return prj_vwu

    def bp(self, prj_vwu: NDArray) -> NDArray:
        """
        Back-projection of the projection data to the volume.

        Parameters
        ----------
        prj_vwu : NDArray
            The projection data to back-project.

        Returns
        -------
        NDArray
            The back-projected volume.
        """
        if self.prj_intensities is not None:
            prj_vwu = prj_vwu * self.prj_intensities[:, np.newaxis]  # Needs copy
        if self.psf_vwu is not None:
            prj_vwu = self.psf_vwu.T(prj_vwu)
        return self.projector_backend.bp(prj_vwu)


class ProjectorAttenuationXRF(ProjectorUncorrected):
    """
    Attenuation corrected projection class for XRF, with multi-detector support.

    It includes the computation of the attenuation volumes.
    """

    att_vol_angles: NDArray[np.floating]

    executor: Union[cf.ThreadPoolExecutor, None]

    def __init__(
        self,
        vol_geom: Union[Sequence[int], NDArray[np.integer], models.VolumeGeometry],
        angles_rot_rad: Union[Sequence[float], NDArray],
        rot_axis_shift_pix: Union[float, ArrayLike, NDArray, None] = None,
        *,
        prj_geom: Optional[models.ProjectionGeometry] = None,
        prj_intensities: Optional[ArrayLike] = None,
        backend: Union[str, prj_backends.ProjectorBackend] = "astra" if astra_available else "skimage",
        att_maps: Optional[NDArray[np.floating]] = None,
        att_in: Optional[NDArray[np.floating]] = None,
        att_out: Optional[NDArray[np.floating]] = None,
        angles_detectors_rad: Union[float, ArrayLike] = (np.pi / 2),
        weights_detectors: Optional[ArrayLike] = None,
        psf: Optional[ArrayLike] = None,
        is_symmetric: bool = False,
        weights_angles: Optional[ArrayLike] = None,
        use_multithreading: bool = True,
        data_type: DTypeLike = np.float32,
        verbose: bool = True,
    ):
        """
        Initialize the (attenuation corrected) XRF dedicated projector.

        Parameters
        ----------
        vol_geom : Sequence[int] | NDArray[np.integer] | models.VolumeGeometry
            The volume shape in X Y and optionally Z.
        angles_rot_rad : Sequence[float] | NDArray
            The rotation angles.
        rot_axis_shift_pix : float | ArrayLike | NDArray | None, optional
            The rotation axis shift(s) in pixels. The default is None.
        prj_geom : Optional[models.ProjectionGeometry], optional
            The fully specified projection geometry.
            When active, the rotation axis shift is ignored. The default is None.
        prj_intensities : Optional[ArrayLike], optional
            Projection scaling factor. The default is None.
        backend : str | prj_backends.ProjectorBackend, optional
            Projector backend to use, by default "astra" if astra is available, otherwise "skimage".
        att_maps : Optional[NDArray[np.floating]], optional
            Precomputed attenuation maps for each angle, by default None
        att_in : Optional[ArrayLike], optional
            Attenuation volume of the incoming beam. The default is None.
        att_out : Optional[ArrayLike], optional
            Attenuation volume of the outgoing beam. The default is None.
        angles_detectors_rad : float | ArrayLike, optional
            Angles of the detector elements with respect to incident beam. The default is (np.pi / 2).
        weights_detectors : Optional[ArrayLike], optional
            Weights (e.g. solid angle, efficiency, etc) of the detector elements. The default is None.
        psf : Optional[ArrayLike], optional
            Optical system's point spread function (PSF). The default is None.
        is_symmetric : bool, optional
            Whether the projector is symmetric or not. The default is False.
        weights_angles : Optional[ArrayLike], optional
            Projection weight for a given element at a given angle. The default is None.
        use_multithreading : bool, optional
            Whether to use multiple threads or not. The default is True.
        data_type : DTypeLike, optional
            Output data type. The default is np.float32.
        verbose : bool, optional
            Whether to produce verbose output. The default is True.

        Raises
        ------
        ValueError
            When given inconsistent numbers of detector weights and detector angles.
        """
        ProjectorUncorrected.__init__(
            self,
            vol_geom=vol_geom,
            angles_rot_rad=angles_rot_rad,
            rot_axis_shift_pix=rot_axis_shift_pix,
            prj_geom=prj_geom,
            psf=psf,
            prj_intensities=prj_intensities,
            backend=backend,
            create_single_projs=True,
        )

        self.data_type = data_type
        self.use_multithreading = use_multithreading
        self.verbose = verbose
        self.is_symmetric = is_symmetric

        self.angles_det_rad = np.array(angles_detectors_rad, ndmin=1)
        num_det_angles = len(self.angles_det_rad)
        if weights_detectors is None:
            weights_detectors = np.ones_like(self.angles_det_rad)
        self.weights_det = np.array(weights_detectors, ndmin=1)

        num_det_weights = len(self.weights_det)
        if num_det_angles > 1:
            if num_det_weights == 1:
                self.weights_det = np.tile(self.weights_det, [num_det_angles])
            elif num_det_weights > 1 and not num_det_weights == num_det_angles:
                raise ValueError(
                    "Number of detector weights differs from number of"
                    f" detector angles: {num_det_weights} vs {num_det_angles}"
                )

        if weights_angles is None:
            weights_angles = np.ones((len(self.angles_rot_rad), num_det_angles))
        else:
            weights_angles = np.array(weights_angles)
        self.weights_angles = weights_angles

        if att_maps is None:
            m = AttenuationVolume(att_in, att_out, self.angles_rot_rad, self.angles_det_rad)
            m.compute_maps(use_multithreading=self.use_multithreading, verbose=self.verbose)
            self.att_vol_angles = m.get_maps()
        else:
            if att_maps.shape[1] != num_det_angles:
                raise ValueError(f"Number of maps ({att_maps.shape[1]}) differs from number of detectors ({num_det_angles})")
            if not np.all(np.equal(att_maps.shape[-2:], self.vol_shape[-2:])):
                raise ValueError(
                    f"Mismatching volume shape of input volume ({att_maps.shape[-2:]})"
                    + f" with vol_shape ({self.vol_shape} in 2D -> {self.vol_shape[-2:]})"
                )
            self.att_vol_angles = att_maps

        self.executor = None

    def __enter__(self):
        """Initialize the with statement block."""
        if self.use_multithreading and isinstance(self.projector_backend, prj_backends.ProjectorBackendSKimage):
            if self.executor is None:
                self.executor = cf.ThreadPoolExecutor(max_workers=num_threads)
        return super().__enter__()

    def __exit__(self, *args):
        """De-initialize the with statement block."""
        super().__exit__()
        if self.use_multithreading and isinstance(self.projector_backend, prj_backends.ProjectorBackendSKimage):
            if self.executor is not None:
                self.executor.shutdown()
                self.executor = None

    def collapse_detectors(self) -> None:
        """Convert multi-detector configurations into single-detector."""
        weights = np.reshape(self.weights_det, [1, -1, 1, 1]) / np.sum(self.weights_det)
        self.att_vol_angles = np.sum(self.att_vol_angles * weights, axis=1)

        weights = np.squeeze(weights)
        self.angles_det_rad = np.sum(self.angles_det_rad * weights, keepdims=True)
        self.weights_angles = np.sum(self.weights_angles * weights, axis=1, keepdims=True)
        self.weights_det = np.sum(self.weights_det, keepdims=True)

    def fp_angle(self, vol: NDArray, angle_ind: int) -> NDArray:
        """Forward-project the volume to a single sinogram line.

        It applies the attenuation corrections.

        Parameters
        ----------
        vol : NDArray
            The volume to forward-project.
        angle_ind : int
            The angle index to forward project.

        Returns
        -------
        sino_line : NDArray
            The forward-projected sinogram line.
        """
        temp_vol = vol * self.att_vol_angles[angle_ind, ...]

        weights = self.weights_det * self.weights_angles[angle_ind, :]
        sino_line = [
            weights[ii] * ProjectorUncorrected.fp_angle(self, temp_vol[ii], angle_ind)
            for ii in range(len(self.angles_det_rad))
        ]
        sino_line = np.ascontiguousarray(np.stack(sino_line, axis=0))

        if sino_line.shape[0] == 1:
            sino_line = np.squeeze(sino_line, axis=0)

        return sino_line

    def bp_angle(self, sino: NDArray, angle_ind: int, single_line: bool = False) -> NDArray:
        """Back-project a single sinogram line to the volume.

        It only applies the attenuation corrections if the projector is symmetric.

        Parameters
        ----------
        sino : NDArray
            The sinogram to back-project or a single line.
        angle_ind : int
            The angle index to forward project.
        single_line : bool, optional
            Whether the input is a single sinogram line. The default is False.

        Returns
        -------
        NDArray
            The back-projected volume.
        """
        if single_line:
            sino_line = sino
        else:
            sino_line = sino[..., angle_ind, :]

        sino_line = np.reshape(sino_line, [len(self.weights_det), *sino_line.shape[-(len(self.vol_shape) - 1) :]])
        weights = self.weights_det * self.weights_angles[angle_ind, :]
        vol = [
            weights[ii] * ProjectorUncorrected.bp_angle(self, sino_line[ii, ...], angle_ind)
            for ii in range(len(self.angles_det_rad))
        ]
        vol = np.stack(vol, axis=0)

        if self.is_symmetric:
            vol *= self.att_vol_angles[angle_ind, ...]

        return np.sum(vol, axis=0)

    def fp(self, vol: NDArray) -> NDArray:
        """Forward-project the volume to the sinogram.

        It applies the attenuation corrections.

        Parameters
        ----------
        vol : NDArray
            The volume to forward-project.

        Returns
        -------
        NDArray
            The forward-projected sinogram.
        """
        if self.use_multithreading and isinstance(self.projector_backend, prj_backends.ProjectorBackendSKimage):
            sino_lines = self.executor.map(lambda x: self.fp_angle(vol, x), range(len(self.angles_rot_rad)))
            sino_lines = [*sino_lines]
        else:
            sino_lines = [self.fp_angle(vol, ii) for ii in range(len(self.angles_rot_rad))]

        sino_lines = np.ascontiguousarray(np.stack(sino_lines, axis=-2))

        if sino_lines.shape[0] == 1:
            sino_lines = np.squeeze(sino_lines, axis=0)

        return sino_lines

    def bp(self, sino: NDArray) -> NDArray:
        """Back-projection of the sinogram to the volume.

        Parameters
        ----------
        sino : NDArray
            The sinogram to back-project.

        Returns
        -------
        NDArray
            The back-projected volume.
        """
        if self.is_symmetric:
            if self.use_multithreading and isinstance(self.projector_backend, prj_backends.ProjectorBackendSKimage):
                vols = self.executor.map(lambda x: self.bp_angle(sino, x), range(len(self.angles_rot_rad)))
                vols = [*vols]
            else:
                vols = [self.bp_angle(sino, ii) for ii in range(len(self.angles_rot_rad))]

            return np.sum(vols, axis=0)
        else:
            sino = np.reshape(sino, [len(self.weights_det), *self.prj_shape])
            vol = [ProjectorUncorrected.bp(self, sino[ii, ...]) for ii in range(len(self.angles_det_rad))]
            return np.sum(np.stack(vol, axis=0), axis=0)
